from insights.plugins.ps_rule_fakes import psaux_no_filter, psauxww_ds_filter, psalxwww_parser_filter
from insights.specs import Specs
from . import InputData, run_test

import pytest


def test_run_test_missing_filters_exception():
    """
    The rule underlying datasource requires a filter,
    an exception should be raised because filter was not
    added in the rule module.
    """
    input_data = InputData("fake_input")
    input_data.add(Specs.ps_aux, "FAKE_CONTENT")
    with pytest.raises(Exception):
        run_test(psaux_no_filter, input_data, None)


def test_run_test_no_missing_filters_using_datasource():
    """
    Required filter was added directly to the datasouce,
    ``run_test`` should complete without any exceptions.
    """
    input_data = InputData("fake_input")
    input_data.add(Specs.ps_auxww, "FAKE_CONTENT")
    result = run_test(psauxww_ds_filter, input_data, None)
    assert result


def test_run_test_no_missing_filters_using_parser():
    """
    Required filter was added to using the parser,
    ``run_test`` should complete without any exceptions.
    """
    input_data = InputData("fake_input")
    input_data.add(Specs.ps_alxwww, "FAKE_CONTENT")
    result = run_test(psalxwww_parser_filter, input_data, None)
    assert result



"""
Uptime
======

Combiner for uptime information. It uses the results of the ``Uptime``
parser to get the uptime information.  ``Uptime`` is
the preferred source of data.

Examples:
    >>> ut = shared[uptime]
    >>> ut.updays
    21
    >>> ht.uptime
    03:45

"""
from __future__ import division
from collections import namedtuple
from insights import SkipComponent
from insights.core.plugins import combiner
from insights.parsers.uptime import Uptime as upt
from insights.util import deprecated

Uptime = namedtuple("Uptime",
                    field_names=[
                        "currtime", "updays", "uphhmm",
                        "users", "loadavg", "uptime"])
"""namedtuple: Type for storing the uptime information."""


@combiner(upt)
def uptime(ut):
    """
    .. warning::
       This combiner method is deprecated, please use
       :py:class:`insights.parsers.uptime.Uptime` instead.

    Check uptime to get the uptime information.

    Returns:
        insights.combiners.uptime.Uptime: A named tuple with `currtime`,
        `updays`, `uphhmm`, `users`, `loadavg` and `uptime` components.

    Raises:
        SkipComponent: If no data is available or if ``loadavg`` is not available.
    """
    deprecated(uptime, "Use the `Uptime` parser instead.", "3.0.300")

    if ut.loadavg:
        return Uptime(ut.currtime, ut.updays, ut.uphhmm,
                      ut.users, ut.loadavg, ut.uptime)

    raise SkipComponent("Unable to get uptime information.")



"""
VDSMId - file ``/etc/vdsm/vdsm.id``
===================================

Module for parsing the content of file ``vdsm.id``, which is a simple file.

Typical content of "vdsm.id" is::

    # VDSM UUID info
    #
    F7D9D983-6233-45C2-A387-9B0C33CB1306

Examples:
    >>> vd = shared[VDSMId]
    >>> vd.uuid
    "F7D9D983-6233-45C2-A387-9B0C33CB1306"

"""
from .. import Parser, parser
from ..parsers import get_active_lines
from insights.specs import Specs


@parser(Specs.vdsm_id)
class VDSMId(Parser):
    """Class for parsing `vdsm.id` file."""

    def parse_content(self, content):
        """
        Returns the UUID of this Host
        - E.g.: F7D9D983-6233-45C2-A387-9B0C33CB1306
        """
        lines = get_active_lines(content)
        self.uuid = lines[0].strip() if len(lines) > 0 else None



from __future__ import print_function
import pytest
import doctest

from insights.parsers import ParseException, SkipException
from insights.parsers import lvm
from insights.tests import context_wrap
from .lvm_test_data import LVMCONFIG, LVMCONFIG2, LVMCONFIG3

WARNINGS_CONTENT = """
WARNING
valid data 1
Checksum Error
valid data 2
  Failed to write
  Attempt To Close Device
valid data 3
""".strip()

WARNINGS_FOUND = """
WARNING
Checksum Error
  Failed to write
  Attempt To Close Device
""".strip()

VGSHEADING_CONTENT = """
Configuration setting "activation/thin_check_executable" unknown.
Configuration setting "activation/thin_check_options" unknown.
Configuration setting "activation/thin_check_executable" unknown.
Configuration setting "activation/thin_check_options" unknown.
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
  Using volume group(s) on command line.
Found duplicate PV qJMs5CEKY89qzq56E2vVeBvxzJGw2sA1: using /dev/mapper/mpathbp2 not /dev/cciss/c0d1p2
Using duplicate PV /dev/mapper/mpathbp2 from subsystem DM, ignoring /dev/cciss/c0d1p2
WARNING: Inconsistent metadata found for VG vgshared - updating to use version 26x
  Archiving volume group "vgshared" metadata (seqno 26).
  Archiving volume group "vgshared" metadata (seqno 27).
  Creating volume group backup "/etc/lvm/backup/vgshared" (seqno 27).
VG       Attr   Ext     #PV #LV #SN VSize   VFree   VG UUID                                VProfile #VMda VMdaFree  VMdaSize  #VMdaUse VG Tags
vg00     wz--n- 128.00m   1   7   0 279.00g   5.00g MiuKdK-ruw4-UG1i-0uLE-4dfB-bk2v-uxte2W              1   506.00k  1020.00k        1
vg00alt  wz--n-   4.00m   1   5   0 279.11g 265.11g LNzjbn-HU4C-WFCM-cRZ6-81ns-4rke-X8N0DA              1   507.00k  1020.00k        1
vg01     wz--n-   4.00m   2   1   0 199.99g 100.99g 7O9ePI-M1Kp-H9VH-Lxt3-pS1Z-nLQB-P1IIO3              2   508.00k  1020.00k        2
vgshared wz--nc   4.00m   6   1   0   1.76t 899.98g fFYYw3-Ns8Q-akPI-rrFU-YYn5-nKkk-UW67qO              6   507.00k  1020.00k        6
  Reloading config files
  Wiping internal VG cache
Configuration setting "activation/thin_check_executable" unknown.
Configuration setting "activation/thin_check_options" unknown.
""".strip()

VGSHEADING_CONTENT_DOC = """
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
Using volume group(s) on command line.
VG            Attr   Ext   #PV #LV #SN VSize   VFree    VG UUID                                VProfile #VMda VMdaFree  VMdaSize  #VMdaUse VG Tags
DATA_OTM_VG   wz--n- 4.00m   6   1   0   2.05t 1020.00m xK6HXk-xl2O-cqW5-2izb-LI9M-4fV0-dAzfcc              6   507.00k  1020.00k        6
ITM_VG        wz--n- 4.00m   1   1   0  16.00g    4.00m nws5dd-INe6-1db6-9U1N-F0G3-S1z2-5XTdO4              1   508.00k  1020.00k        1
ORABIN_OTM_VG wz--n- 4.00m   2   3   0 190.00g       0  hfJwg8-hset-YgUY-X6NJ-gkWE-EunZ-KuCXGP              2   507.50k  1020.00k        2
REDO_OTM_VG   wz--n- 4.00m   1   3   0  50.00g       0  Q2YtGy-CWKU-sEYj-mqHk-rbdP-Hzup-wi8jsf              1   507.50k  1020.00k        1
SWAP_OTM_VG   wz--n- 4.00m   1   1   0  24.00g    8.00g hAerzZ-U8QU-ICkc-xxCj-N2Ny-rWzq-pmTpWJ              1   508.00k  1020.00k        1
rootvg        wz--n- 4.00m   1   6   0  19.51g    1.95g p4tLLb-ikeo-Ankk-2xJ6-iHYf-D4E6-KFCFvr              1   506.50k  1020.00k        1
Reloading config files
Wiping internal VG cache
""".strip()

LVS_HAEADING_OUTUPT = """
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
LV          VG      Attr       LSize  Pool Origin Data%  Meta%  Move Log Cpy%Sync Convert LV Tags Devices
lv_app      vg_root -wi-ao---- 71.63g                                                             /dev/sda2(7136)
lv_home     vg_root -wi-ao----  2.00g                                                             /dev/sda2(2272)
lv_opt      vg_root -wi-ao----  5.00g                                                             /dev/sda2(2784)
lv_root     vg_root -wi-ao----  5.00g                                                             /dev/sda2(0)
lv_tmp      vg_root -wi-ao----  1.00g                                                             /dev/sda2(4064)
lv_usr      vg_root -wi-ao----  5.00g                                                             /dev/sda2(4320)
lv_usrlocal vg_root -wi-ao----  1.00g                                                             /dev/sda2(5600)
lv_var      vg_root -wi-ao----  5.00g                                                             /dev/sda2(5856)
swap        vg_root -wi-ao----  3.88g                                                             /dev/sda2(1280)
""".strip()

CONTENT_WITH_EXTRA_LOCK_TIPS = """
WARNING: locking_type (0) is deprecated, using --nolocking.
WARNING: File locking is disabled.
Reading VG shared_vg1 without a lock.
LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='V1AMDw-e9x6-AdCq-p3Hr-6cYn-otra-ducLAc'|LVM2_VG_NAME='rhel'|LVM2_VG_ATTR='wz--n-'|LVM2_VG_PERMISSIONS='writeable'|LVM2_VG_EXTENDABLE='extendable'|LVM2_VG_EXPORTED=''|LVM2_VG_AUTOACTIVATION='enabled'|LVM2_VG_PARTIAL=''|LVM2_VG_ALLOCATION_POLICY='normal'|LVM2_VG_CLUSTERED=''|LVM2_VG_SHARED=''|LVM2_VG_SIZE='<29.00g'|LVM2_VG_FREE='0 '|LVM2_VG_SYSID=''|LVM2_VG_SYSTEMID=''|LVM2_VG_LOCK_TYPE=''|LVM2_VG_LOCK_ARGS=''|LVM2_VG_EXTENT_SIZE='4.00m'|LVM2_VG_EXTENT_COUNT='7423'|LVM2_VG_FREE_COUNT='0'|LVM2_MAX_LV='0'|LVM2_MAX_PV='0'|LVM2_PV_COUNT='1'|LVM2_VG_MISSING_PV_COUNT='0'|LVM2_LV_COUNT='2'|LVM2_SNAP_COUNT='0'|LVM2_VG_SEQNO='3'|LVM2_VG_TAGS=''|LVM2_VG_PROFILE=''|LVM2_VG_MDA_COUNT='1'|LVM2_VG_MDA_USED_COUNT='1'|LVM2_VG_MDA_FREE='507.50k'|LVM2_VG_MDA_SIZE='1020.00k'|LVM2_VG_MDA_COPIES='unmanaged'
"""

LVM_CONF = """
locking_type = 1
#locking_type = 2
# volume_list = [ "vg1", "vg2/lvol1", "@tag1", "@*" ]
volume_list = [ "vg2", "vg3/lvol3", "@tag2", "@*" ]
# filter = [ "a|loop|", "r|/dev/hdc|", "a|/dev/ide|", "r|.*|" ]

filter = [ "r/sda[0-9]*$/",  "a/sd.*/" ]
filter = [ "a/sda[0-9]*$/",  "r/sd.*/" ]
shell {
    history_size = 100
}
""".strip()

PVS_HEADINGS_OUTPUT = """
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
Scanning all devices to update lvmetad.
No PV label found on /dev/loop0.
No PV label found on /dev/loop1.
No PV label found on /dev/sda1.
No PV label found on /dev/fedora/root.
No PV label found on /dev/sda2.
No PV label found on /dev/fedora/swap.
No PV label found on /dev/fedora/home.
No PV label found on /dev/mapper/docker-253:1-2361272-pool.
Wiping internal VG cache
Wiping cache of LVM-capable devices
PV                                                    VG     Fmt  Attr PSize   PFree DevSize PV UUID                                PMdaFree  PMdaSize  #PMda #PMdaUse PE
/dev/fedora/home                                                  ---       0     0  418.75g                                               0         0      0        0      0
/dev/fedora/root                                                  ---       0     0   50.00g                                               0         0      0        0      0
/dev/fedora/swap                                                  ---       0     0    7.69g                                               0         0      0        0      0
/dev/loop0                                                        ---       0     0  100.00g                                               0         0      0        0      0
/dev/loop1                                                        ---       0     0    2.00g                                               0         0      0        0      0
/dev/mapper/docker-253:1-2361272-pool                             ---       0     0  100.00g                                               0         0      0        0      0
/dev/mapper/luks-7430952e-7101-4716-9b46-786ce4684f8d fedora lvm2 a--  476.45g 4.00m 476.45g FPLCRf-d918-LVL7-6e3d-n3ED-aiZv-EesuzY        0   1020.00k     1        1 121970
/dev/sda1                                                         ---       0     0  500.00m                                               0         0      0        0      0
/dev/sda2                                                         ---       0     0  476.45g                                               0         0      0        0      0
Reloading config files
Wiping internal VG cache
""".strip()

VGS_OUTPUT = """
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
Using volume group(s) on command line.
VG            Attr   Ext   #PV #LV #SN VSize   VFree    VG UUID                                VProfile #VMda VMdaFree  VMdaSize  #VMdaUse VG Tags
DATA_OTM_VG   wz--n- 4.00m   6   1   0   2.05t 1020.00m xK6HXk-xl2O-cqW5-2izb-LI9M-4fV0-dAzfcc              6   507.00k  1020.00k        6
ITM_VG        wz--n- 4.00m   1   1   0  16.00g    4.00m nws5dd-INe6-1db6-9U1N-F0G3-S1z2-5XTdO4              1   508.00k  1020.00k        1
ORABIN_OTM_VG wz--n- 4.00m   2   3   0 190.00g       0  hfJwg8-hset-YgUY-X6NJ-gkWE-EunZ-KuCXGP              2   507.50k  1020.00k        2
REDO_OTM_VG   wz--n- 4.00m   1   3   0  50.00g       0  Q2YtGy-CWKU-sEYj-mqHk-rbdP-Hzup-wi8jsf              1   507.50k  1020.00k        1
SWAP_OTM_VG   wz--n- 4.00m   1   1   0  24.00g    8.00g hAerzZ-U8QU-ICkc-xxCj-N2Ny-rWzq-pmTpWJ              1   508.00k  1020.00k        1
rootvg        wz--n- 4.00m   1   6   0  19.51g    1.95g p4tLLb-ikeo-Ankk-2xJ6-iHYf-D4E6-KFCFvr              1   506.50k  1020.00k        1
Reloading config files
Wiping internal VG cache
""".strip()

SYSTEM_DEVICES1 = """
# LVM uses devices listed in this file.
# Created by LVM command lvmdevices pid 2631 at Fri May 27 07:37:11 2022
VERSION=1.1.2
IDTYPE=devname IDNAME=/dev/vda2 DEVNAME=/dev/vda2 PVID=phl0clFbAokp9UXqbIgI5YYQxuTIJVkD PART=2
""".strip()

SYSTEM_DEVICES2 = """
# LVM uses devices listed in this file.
# Created by LVM command lvmdevices pid 2631 at Fri May 27 07:37:11 2022
VERSION=1.1.2
IDTYPE=sys_wwid IDNAME=/dev/vda2 DEVNAME=/dev/vda2 PVID=phl0clFbAokp9UXqbIgI5YYQxuTIJVkD PART=2
IDTYPE=sys_serial IDNAME=/dev/vda1 DEVNAME=/dev/vda1 PVID=phl0clFbAokp9UXqbIgI5YYQdeTIJVkD PART=1
""".strip()

SYSTEM_DEVICES3 = """
# LVM uses devices listed in this file.
# Created by LVM command lvmdevices pid 2631 at Fri May 27 07:37:11 2022
VERSION=1.1.2
""".strip()


def test_find_warnings():
    data = [l for l in lvm.find_warnings(WARNINGS_CONTENT.splitlines())]
    assert len(data) == len(WARNINGS_FOUND.splitlines())
    assert data == WARNINGS_FOUND.splitlines()


def compare_partial_dicts(result, expected):
    """
    Make sure all the keys in expected are matched by keys in result, and
    that the values stored in those keys match.  Result can contain more
    items than expected - those are ignored.

    Used in the test_lvs, test_pvs and test_vgs tests.
    """
    # return all(result[k] == expected[k] for k in expected.keys())
    mismatches = 0
    for k in expected.keys():
        if not result[k] == expected[k]:
            print("Failed for key {k}, {r} != {e}".format(k=k, r=result[k], e=expected[k]))
            mismatches += 1
    return mismatches == 0


def test_lvmconfig():
    p = lvm.LvmConfig(context_wrap(LVMCONFIG))
    assert p.data["dmeventd"]["raid_library"] == "libdevmapper-event-lvm2raid.so"
    assert p.data["global"]["thin_check_options"] == ["-q", "--clear-needs-check-flag"]


def test_lvmconfig_2():
    p = lvm.LvmConfig(context_wrap(LVMCONFIG2))
    assert p.data['global']['umask'] == '077'


def test_lvmconfig_exception():
    with pytest.raises(ParseException):
        lvm.LvmConfig(context_wrap(LVMCONFIG3))


def test_vgsheading_warnings():
    result = lvm.VgsHeadings(context_wrap(VGSHEADING_CONTENT))
    assert len(result.warnings) == 6
    assert 'Configuration setting "activation/thin_check_executable" unknown.' in result.warnings
    assert 'WARNING: Locking disabled. Be careful! This could corrupt your metadata.' in result.warnings


def test_vgs_with_extra_tips():
    result = lvm.Vgs(context_wrap(CONTENT_WITH_EXTRA_LOCK_TIPS))
    assert result is not None
    assert len(result.data.get('warnings')) == 3
    assert 'Reading VG shared_vg1 without a lock.' in result.data.get('warnings')
    assert len(result.data.get('content')) == 1


def test_system_devices():
    devices = lvm.LvmSystemDevices(context_wrap(SYSTEM_DEVICES2))
    assert len(devices) == 2
    assert '/dev/vda1' in devices
    assert devices['/dev/vda1']['IDTYPE'] == 'sys_serial'
    assert '/dev/vda2' in devices
    assert devices['/dev/vda2']['IDTYPE'] == 'sys_wwid'


def test_system_devices_exception():
    with pytest.raises(SkipException):
        lvm.LvmSystemDevices(context_wrap(SYSTEM_DEVICES3))


def test_docs():
    env = {
        'devices': lvm.LvmSystemDevices(context_wrap(SYSTEM_DEVICES1)),
        'lvm_conf_data': lvm.LvmConf(context_wrap(LVM_CONF)),
        'pvs_data': lvm.PvsHeadings(context_wrap(PVS_HEADINGS_OUTPUT)),
        'vgs_info': lvm.VgsHeadings(context_wrap(VGSHEADING_CONTENT_DOC)),
        'lvs_info': lvm.LvsHeadings(context_wrap(LVS_HAEADING_OUTUPT))
    }
    failed, total = doctest.testmod(lvm, globs=env)
    assert failed == 0



from insights.parsers import yum_updates
from insights.tests import context_wrap
import doctest

YUM_UPDATES_INPUT = """{
      "releasever": "8",
      "basearch": "x86_64",
      "update_list": {
        "NetworkManager-1:1.22.8-4.el8.x86_64": {
          "available_updates": [
            {
              "package": "NetworkManager-1:1.22.8-5.el8_2.x86_64",
              "repository": "rhel-8-for-x86_64-baseos-rpms",
              "basearch": "x86_64",
              "releasever": "8",
              "erratum": "RHSA-2020:3011"
            }
          ]
        }
      },
      "metadata_time": "2021-01-01T09:39:45Z"
    }
"""


def test_yum_updates():
    info = yum_updates.YumUpdates(context_wrap(YUM_UPDATES_INPUT))
    assert info is not None
    assert len(info.updates) == 1
    assert len(info.updates["NetworkManager-1:1.22.8-4.el8.x86_64"]["available_updates"]) == 1


def test_yum_updates_docs():
    env = {
        'updates': yum_updates.YumUpdates(context_wrap(YUM_UPDATES_INPUT)),
    }
    failed, total = doctest.testmod(yum_updates, globs=env)
    assert failed == 0



"""
Manila configuration - file ``/etc/manila/manila.conf``
=======================================================
"""
from insights.core import IniConfigFile
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.manila_conf)
class ManilaConf(IniConfigFile):
    """
    The Manila configuration file is a standard '.ini' file and this parser uses
    the ``IniConfigFile`` class to read it.

    Sample configuration::

        [DEFAULT]
        osapi_max_limit = 1000
        use_forwarded_for = false
        api_paste_config = api-paste.ini
        state_path = /var/lib/manila
        scheduler_topic = manila-scheduler
        share_topic = manila-share
        share_driver = manila.share.drivers.generic.GenericShareDriver
        enable_v1_api = false
        enable_v2_api = false

        [cors]
        allow_credentials = true
        expose_headers = Content-Type,Cache-Control,Content-Language,Expires,Last-Modified,Pragma
        allow_methods = GET,POST,PUT,DELETE,OPTIONS
        allow_headers = Content-Type,Cache-Control,Content-Language,Expires,Last-Modified,Pragma

    Examples:
        >>> type(conf)
        <class 'insights.parsers.manila_conf.ManilaConf'>
        >>> conf.sections()
        ['cors']
        >>> 'cors' in conf
        True
        >>> conf.has_option('DEFAULT', 'share_topic')
        True
        >>> conf.get("DEFAULT", "share_topic")
        'manila-share'
        >>> conf.get("DEFAULT", "enable_v2_api")
        'false'
        >>> conf.getboolean("DEFAULT", "enable_v2_api")
        False
        >>> conf.getint("DEFAULT", "osapi_max_limit")
        1000
    """
    pass



#!/usr/bin/env python
"""
Example Cluster Rule - bash_version
===================================

This rule can be run from the root directory of the insights-core
repo using the following command line::

    $ insights-run -p examples.cluster_rules.bash_version \
            examples/cluster_rules/cluster_hosts.tar.gz

or from the examples/cluster_rules directory::

    $ ./bash_version.py cluster_hosts.tar.gz

"""
from insights import run, make_pass, make_fail
from insights.parsers.installed_rpms import InstalledRpms
from insights.combiners.hostname import hostname
from insights.core.plugins import fact, rule


@fact(InstalledRpms)
def bash_version(rpms):
    """ dict: Returns rpm info for bash on each cluster host """
    rpm = rpms.get_max("bash")
    return {"name": rpm.name, "version": rpm.nvr}


@fact(hostname)
def get_hostname(host):
    """ dict: Returns hostname info for each cluster host """
    return {"hostname": host.fqdn}


@rule(bash_version, get_hostname, cluster=True)
def bash_rule(bash, hostnames):
    """
    Cluster rule to process bash and hostname info

    ``bash`` and ``hostnames`` are Pandas DataFrames for the facts collected
    for each host in the cluster.  See
    https://pandas.pydata.org/pandas-docs/stable/api.html#dataframe
    for information on available attributes and methods.

    Arguments:
        bash (pandas.DataFrame): Includes facts from ``bash_version``
            fact with columns "name" and "version" and one row per
            host in the cluster.
        hostnames (pandas.DataFrame): Includes facts from ``get_hostname``
            fact with column "hostname" and one row per
            host in the cluster.
    """
    if isinstance(bash, dict):
        return make_fail('bash_rule',
                         error_message="Run this rule with a cluster archive")

    return make_pass('bash_rule', bash=bash, hostname=hostnames)


if __name__ == "__main__":
    run(bash_rule, print_summary=True)



from insights.parsers.wc_proc_1_mountinfo import WcProc1Mountinfo
from insights.parsers import wc_proc_1_mountinfo
from insights.tests import context_wrap
from insights.core.dr import SkipComponent
from insights.parsers import ParseException
import pytest
import doctest

WC_PROC_1_MOUNTINFO_ERR1 = """
-bash: /usr/bin/wcq: No such file or directory
""".strip()

WC_PROC_1_MOUNTINFO_ERR2 = """
/usr/bin/wc: /proc/100/mountinfo: No such file or directory
""".strip()

WC_PROC_1_MOUNTINFO_ERR3 = """
unknow /proc/1/mountinfo
""".strip()

WC_PROC_1_MOUNTINFO = """
37 /proc/1/mountinfo
""".strip()


def test_wc_proc_1_mountinfo():
    results = WcProc1Mountinfo(context_wrap(WC_PROC_1_MOUNTINFO))
    assert results.line_count == 37


def test_wc_proc_1_mountinfo_errors():
    with pytest.raises(SkipComponent) as pe:
        WcProc1Mountinfo(context_wrap(WC_PROC_1_MOUNTINFO_ERR1))
        assert WC_PROC_1_MOUNTINFO_ERR1 in str(pe)

    with pytest.raises(SkipComponent) as pe:
        WcProc1Mountinfo(context_wrap(WC_PROC_1_MOUNTINFO_ERR2))
        assert WC_PROC_1_MOUNTINFO_ERR2 in str(pe)

    with pytest.raises(ParseException) as pe:
        WcProc1Mountinfo(context_wrap(WC_PROC_1_MOUNTINFO_ERR3))
        assert WC_PROC_1_MOUNTINFO_ERR3 in str(pe)


def test_doc_examples():
    env = {
            'wc_info': WcProc1Mountinfo(context_wrap(WC_PROC_1_MOUNTINFO)),
          }
    failed, total = doctest.testmod(wc_proc_1_mountinfo, globs=env)
    assert failed == 0



"""
SSSDLog - files matching ``/var/log/sssd/*.log``
================================================

"""

from .. import parser, LogFileOutput
from datetime import datetime
from insights.specs import Specs


def strip_surrounds(s):
    # Or s{([\[\(]+)(.*)\1:?}{\2} except matching brace style
    start = 0
    end = len(s) - 1
    if s[end] == ':':
        end -= 1
    # print "strip '{s}' from {start} to {end}".format(s=s, start=start, end=end)
    while (s[start] == '[' and s[end] == ']') or \
      (s[start] == '(' and s[end] == ')'):
        start += 1
        end -= 1
    return s[start:(end + 1)]


@parser(Specs.sssd_logs)
class SSSDLog(LogFileOutput):
    """
    Parser class for reading SSSD log files.  The main work is done by the
    LogFileOutput super-class.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`

    Sample input::

        (Tue Feb 14 09:45:02 2017) [sssd] [sbus_remove_timeout] (0x2000): 0x7f5aceb6a970
        (Tue Feb 14 09:45:02 2017) [sssd] [sbus_dispatch] (0x4000): dbus conn: 0x7f5aceb5cff0
        (Tue Feb 14 09:45:02 2017) [sssd] [sbus_dispatch] (0x4000): Dispatching.
        (Tue Feb 14 09:45:02 2017) [sssd] [sbus_remove_timeout] (0x2000): 0x7f5aceb63eb0
        (Tue Feb 14 09:45:02 2017) [sssd] [sbus_dispatch] (0x4000): dbus conn: 0x7f5aceb578b0
        (Tue Feb 14 09:45:02 2017) [sssd] [sbus_dispatch] (0x4000): Dispatching.
        (Tue Feb 14 09:45:02 2017) [sssd] [sbus_remove_timeout] (0x2000): 0x7f5aceb60f30
        (Tue Feb 14 09:45:02 2017) [sssd] [sbus_dispatch] (0x4000): dbus conn: 0x7f5aceb58360
        (Tue Feb 14 09:45:06 2015) [sssd] [monitor_hup] (0x0020): Received SIGHUP.
        (Tue Feb 14 09:45:07 2015) [sssd] [te_server_hup] (0x0020): Received SIGHUP. Rotating logfiles.

    Each line is parsed into a dictionary with the following keys:

        * **timestamp** - the date of the log line (as a string)
        * **datetime** - the date as a datetime object (if conversion is possible)
        * **module** - the module logging the message
        * **function** - the function within the module
        * **level** - the debug level (as a string)
        * **message** - the body of the message
        * **raw_message** - the raw message before being split.

    Examples:

        >>> logs = shared[SSSDLog]

        >>> hups = logs.get("SIGHUP")
        >>> print len(hups)
        2
        >>> hups[0]['module']
        'monitor_hup'

    """
    time_format = '%b %d %H:%M:%S %Y'

    def _parse_line(self, line):
        fields = line.split()
        parsed_line = {
            'timestamp': strip_surrounds(' '.join(fields[0:5])),
            'module': strip_surrounds(fields[5]),
            'function': strip_surrounds(fields[6]),
            'level': strip_surrounds(fields[7]),
            'message': ' '.join(fields[8:]),
            'raw_message': line
        }
        # Try to convert the datetime if possible
        try:
            parsed_line['datetime'] = datetime.strptime(
                parsed_line['timestamp'],
                '%a %b %d %H:%M:%S %Y'
            )
        except:
            pass
        return parsed_line



"""
SCTP Socket State Parser
========================

Parsers provided by this module include:

SCTPEps - file ``/proc/net/sctp/eps``
-------------------------------------

SCTPAsc - file ``/proc/net/sctp/assocs`` on RHEL-6
---------------------------------------------------

SCTPAsc7 - file ``/proc/net/sctp/assocs`` on RHEL-7
---------------------------------------------------

SCTPSnmp - file ``/proc/net/sctp/snmp``
---------------------------------------
"""

from insights import Parser, parser
from insights.components.rhel_version import IsRhel6, IsRhel7
from insights.parsers import keyword_search, SkipException, ParseException
from insights.specs import Specs


@parser(Specs.sctp_eps)
class SCTPEps(Parser):
    """
    This parser parses the content of ``/proc/net/sctp/eps`` file.
    It returns a list of dictionaries. The dictionary contains detail
    information of individual SCTP endpoint, which includes Endpoints, Socket,
    Socket type, Socket State, hash bucket, bind port, UID, socket inodes,
    Local IP address.

    Typical contents of ``/proc/net/sctp/eps`` file are::

        ENDPT            SOCK             STY SST HBKT LPORT   UID INODE     LADDRS
        ffff88017e0a0200 ffff880300f7fa00 2   10  29   11165   200 299689357 10.0.0.102 10.0.0.70
        ffff880612e81c00 ffff8803c28a1b00 2   10  30   11166   200 273361203 10.0.0.102 10.0.0.70 172.31.1.2

    Output data is stored in the list of dictionaries

    Examples:
        >>> type(sctp_info)
        <class 'insights.parsers.sctp.SCTPEps'>
        >>> sorted(sctp_info.sctp_local_ports) == sorted(['11165', '11166'])
        True
        >>> sorted(sctp_info.sctp_local_ips) == sorted(['10.0.0.102', '10.0.0.70', '172.31.1.2'])
        True
        >>> sorted(sctp_info.search(local_port="11165")) == sorted([{'endpoints': 'ffff88017e0a0200', 'socket': 'ffff880299f7fa00', 'sk_type': '2', 'sk_state': '10', 'hash_bkt': '29', 'local_port': '11165', 'uid': '200', 'inode': '299689357', 'local_addr': ['10.0.0.102', '10.0.0.70']}])
        True
        >>> len(sctp_info.search(local_port="11165")) == 1
        True
        >>> len(sctp_info.search(endpoints="ffff88017e0a0200")) == 1
        True
        >>> sctp_info.sctp_eps_ips
        {'ffff88017e0a0200': ['10.0.0.102', '10.0.0.70'], 'ffff880612e81c00': ['10.0.0.102', '10.0.0.70', '172.31.1.2']}
    """

    COLUMN_IDX = [
        'endpoints',
        'socket',
        'sk_type',
        'sk_state',
        'hash_bkt',
        'local_port',
        'uid',
        'inode',
        'local_addr'
    ]

    def parse_content(self, content):
        if (not content) or (not self.file_path):
            raise SkipException("No Contents")

        line = content[0].strip().split()
        keys_cnt = len(self.COLUMN_IDX)
        if "LPORT" not in line or len(line) != keys_cnt:
            raise ParseException("The following line is not compatible with this parser: {0}".format(line))

        self.data = []
        for line in content[1:]:
            line = line.strip().split(None, keys_cnt - 1)
            line[-1] = line[-1].split()
            self.data.append(dict(zip(self.COLUMN_IDX, line)))

        self._sctp_local_ports = set()
        self._sctp_local_ips = set()
        self._sctp_eps_ips = {}
        for line in self.data:
            self._sctp_local_ports.add(line['local_port'])
            local_addr = line['local_addr']
            self._sctp_local_ips.update(local_addr)
            if line['endpoints'] not in self._sctp_eps_ips:
                self._sctp_eps_ips[line['endpoints']] = []
            self._sctp_eps_ips[line['endpoints']].extend(local_addr)

    @property
    def sctp_local_ports(self):
        """
        (list): This function returns a list of SCTP ports if SCTP
                endpoints are created, else `[]`.
        """
        return sorted(self._sctp_local_ports)

    @property
    def sctp_local_ips(self):
        """
        (list): This function returns a list of all local ip addresses
                if SCTP endpoints are created, else `[]`.
        """
        return sorted(self._sctp_local_ips)

    @property
    def sctp_eps_ips(self):
        """
        (dict): This function returns a dict of all endpoints and corresponding
                local ip addresses used by SCTP endpoints if SCTP endpoints are
                created, else `{}`.
        """
        return self._sctp_eps_ips

    def search(self, **args):
        """
        (list): This function return a list of all endpoints when args search matches,
                when args search do not match then it returns `[]`.
        """
        return keyword_search(self.data, **args)


class SCTPAscBase(Parser):
    """
    This parser parses the content of ``/proc/net/sctp/assocs`` file.
    And returns a list of dictionaries. The dictionary contains details
    of individual SCTP endpoint, which includes Association Struct, Socket,
    Socket type, Socket State, Association state, hash bucket, association id,
    tx queue, rx queue, uid, inode, local port, remote port, 'local addr,
    remote addr, heartbeat interval, max in-stream, max out-stream, max
    retransmission attempt, number of init chunks send, number of shutdown
    chunks send, data chunks retransmitted'
    """

    def parse_content(self, content):
        self.data = []
        self._sctp_local_ports = set()
        self._sctp_remote_ports = set()
        self._sctp_local_ips = set()
        self._sctp_remote_ips = set()
        if (not content) or (not self.file_path):
            raise SkipException("No Contents")

        line = content[0].strip().split()
        keys_cnt = len(self.COLUMN_IDX)
        if "LPORT" not in line or len(line) != keys_cnt:
            raise ParseException("The following line is not compatible with this parser: {0}".format(line))

        laddr_idx = line.index('LADDRS')
        raddr_ridx = len(line) - line.index('RADDRS')
        for line in content[1:]:
            line_1 = line.strip().split(None, laddr_idx)
            line_end = line_1.pop()
            idx = line_end.index('<->')
            laddrs = line_end[:idx].strip().split()
            line_end = line_end[idx + 3:].strip().rsplit(None, raddr_ridx - 1)
            raddrs = line_end.pop(0).split()
            line_1.append(laddrs)
            line_1.append(raddrs)
            line_1.extend(line_end)
            self.data.append(dict(zip(self.COLUMN_IDX[:-1], line_1)))

        for line in self.data:
            self._sctp_local_ports.add(line['local_port'])
            self._sctp_remote_ports.add(line['remote_port'])
            self._sctp_local_ips.update(line['local_addr'])
            self._sctp_remote_ips.update(line['remote_addr'])

    @property
    def sctp_local_ports(self):
        """
        (list): This function returns a list of SCTP local peer ports
                if SCTP endpoints are created, else `[]`.
        """
        return sorted(self._sctp_local_ports)

    @property
    def sctp_remote_ports(self):
        """
        (list): This function returns a list of SCTP remote peer ports
                if SCTP endpoints are created, else `[]`.
        """
        return sorted(self._sctp_remote_ports)

    @property
    def sctp_local_ips(self):
        """
        (list): This function returns a list of all local peer's ip addresses
                if SCTP endpoints are created, else `[]`.
        """
        return sorted(self._sctp_local_ips)

    @property
    def sctp_remote_ips(self):
        """
        (list): This function returns a list of all remote peer's ip addresses
                if SCTP endpoints are created, else `[]`.
        """
        return sorted(self._sctp_remote_ips)

    def search(self, **args):
        """
        (list): This function return a list of all SCTP associations when args search matches,
                when args search do not match then it returns `[]`.
        """
        return keyword_search(self.data, **args)


@parser(Specs.sctp_asc, IsRhel6)
class SCTPAsc(SCTPAscBase):
    """
    This parser parses the file ``/proc/net/sctp/assocs`` from RHEL-6. It has
    different columns as compare to RHEL-7.

    Typical contents of ``/proc/net/sctp/assocs`` on RHEL-6 file are::

        ASSOC     SOCK   STY SST ST HBKT ASSOC-ID TX_QUEUE RX_QUEUE UID INODE LPORT RPORT LADDRS <-> RADDRS HBINT INS OUTS MAXRT T1X T2X RTXC
        ffff88045ac7e000 ffff88062077aa00 2   1   4  1205  963        0        0     200 273361167 11567 11166  10.0.0.102 10.0.0.70 <-> *10.0.0.109 10.0.0.77      1000     2     2   10    0    0        0
        ffff88061fbf2000 ffff88060ff92500 2   1   4  1460  942        0        0     200 273360669 11566 11167  10.0.0.102 10.0.0.70 <-> *10.0.0.109 10.0.0.77      1000     2     2   10    0    0        0

    Output data is stored in the list of dictionaries

    Examples:
        >>> type(sctp_asc)
        <class 'insights.parsers.sctp.SCTPAsc'>
        >>> sorted(sctp_asc.sctp_local_ports) == sorted(['11567','11566'])
        True
        >>> sorted(sctp_asc.sctp_remote_ports) == sorted(['11166','11167'])
        True
        >>> sorted(sctp_asc.sctp_local_ips) == sorted(['10.0.0.102', '10.0.0.70'])
        True
        >>> sorted(sctp_asc.sctp_remote_ips) == sorted(['*10.0.0.109', '10.0.0.77'])
        True
        >>> sorted(sctp_asc.search(local_port='11566')) == sorted([{'init_chunks_send': '0', 'uid': '200', 'shutdown_chunks_send': '0', 'max_outstream': '2', 'tx_que': '0', 'inode': '273360669', 'hrtbt_intrvl': '1000', 'sk_type': '2', 'remote_addr': ['*10.0.0.109', '10.0.0.77'], 'data_chunks_retrans': '0', 'local_addr': ['10.0.0.102', '10.0.0.70'], 'asc_id': '942', 'max_instream': '2', 'remote_port': '11167', 'asc_state': '4', 'max_retrans_atmpt': '10', 'sk_state': '1', 'socket': 'ffff88060ff92500', 'asc_struct': 'ffff88061fbf2000', 'local_port': '11566', 'hash_bkt': '1460', 'rx_que': '0'}])
        True
    """

    def __init__(self, *args, **kwargs):
        self.COLUMN_IDX = [
            'asc_struct',
            'socket',
            'sk_type',
            'sk_state',
            'asc_state',
            'hash_bkt',
            'asc_id',
            'tx_que',
            'rx_que',
            'uid',
            'inode',
            'local_port',
            'remote_port',
            'local_addr',
            'remote_addr',
            'hrtbt_intrvl',
            'max_instream',
            'max_outstream',
            'max_retrans_atmpt',
            'init_chunks_send',
            'shutdown_chunks_send',
            'data_chunks_retrans',
            'relation',   # should be ignore
        ]
        super(SCTPAsc, self).__init__(*args, **kwargs)


@parser(Specs.sctp_asc, IsRhel7)
class SCTPAsc7(SCTPAscBase):
    """
    This parser parses the file ``/proc/net/sctp/assocs`` from RHEL-7. It has
    different columns as compare to RHEL-6.

    Typical contents of ``/proc/net/sctp/assocs`` on RHEL-7 file are::

        ASSOC     SOCK   STY SST ST HBKT ASSOC-ID TX_QUEUE RX_QUEUE UID INODE LPORT RPORT LADDRS <-> RADDRS HBINT INS OUTS MAXRT T1X T2X RTXC wmema wmemq sndbuf rcvbuf
        ffff8805d36b3000 ffff880f8911f380 0   10  3  0    12754        0        0       0 496595 3868   3868  10.131.222.5 <-> *10.131.160.81 10.131.176.81        30000    17    10   10    0    0        0        11        12  1000000  2000000
        ffff8805f17e1000 ffff881004aff380 0   10  3  0    12728        0        0       0 532396 3868   3868  10.131.222.3 <-> *10.131.160.81 10.131.176.81        30000    17    10   10    0    0        0        13        14  3000000  4000000


    Output data is stored in the list of dictionaries

    Examples:
        >>> type(sctp_asc_7)
        <class 'insights.parsers.sctp.SCTPAsc7'>
        >>> sctp_asc_7.sctp_local_ips == sorted(['10.131.222.5', '10.131.222.3'])
        True
        >>> sctp_asc_7.data[0]['rcvbuf']
        '2000000'
    """

    def __init__(self, *args, **kwargs):
        self.COLUMN_IDX = [
            'asc_struct',
            'socket',
            'sk_type',
            'sk_state',
            'asc_state',
            'hash_bkt',
            'asc_id',
            'tx_que',
            'rx_que',
            'uid',
            'inode',
            'local_port',
            'remote_port',
            'local_addr',
            'remote_addr',
            'hrtbt_intrvl',
            'max_instream',
            'max_outstream',
            'max_retrans_atmpt',
            'init_chunks_send',
            'shutdown_chunks_send',
            'data_chunks_retrans',
            'wmema',
            'wmemq',
            'sndbuf',
            'rcvbuf',
            'relation',   # should be ignore
        ]
        super(SCTPAsc7, self).__init__(*args, **kwargs)


@parser(Specs.sctp_snmp)
class SCTPSnmp(Parser, dict):
    """
    This parser parses the content of ``/proc/net/sctp/snmp`` file,
    which contains statistics related to SCTP states, packets and chunks.

    Sample content::

        SctpCurrEstab                   	5380
        SctpActiveEstabs                	12749
        SctpPassiveEstabs               	55
        SctpAborteds                    	2142
        SctpShutdowns                   	5295
        SctpOutOfBlues                  	36786
        SctpChecksumErrors              	0
        SctpOutCtrlChunks               	1051492

    Data is stored in a dictionary.

    Examples:

        >>> type(sctp_snmp)
        <class 'insights.parsers.sctp.SCTPSnmp'>
        >>> sctp_snmp.get('SctpCurrEstab')
        5380
        >>> sctp_snmp.get('SctpChecksumErrors') == 0
        True
        >>> 'SctpShutdowns' in sctp_snmp
        True
        >>> len(sctp_snmp)
        8
        >>> sorted(sctp_snmp.keys())
        ['SctpAborteds', 'SctpActiveEstabs', 'SctpChecksumErrors', 'SctpCurrEstab', 'SctpOutCtrlChunks', 'SctpOutOfBlues', 'SctpPassiveEstabs', 'SctpShutdowns']

    Resultant Data::

        {
            'SctpCurrEstab': 5380,
            'SctpActiveEstabs': 12749,
            'SctpPassiveEstabs': 55,
            'SctpAborteds': 2142,
            'SctpShutdowns': 5295,
            'SctpOutOfBlues': 36786,
            'SctpChecksumErrors': 0,
            ...
            ...
        }

    Raises:
        SkipException: When contents are empty.
        ParseException: When file contents are not in expected format.
    """

    def parse_content(self, content):
        if (not content) or (not self.file_path):
            raise SkipException("No Contents")

        for line in content:
            line_strip = line.split()
            if len(line_strip) != 2:
                raise ParseException("Contents are not compatible to this parser")

            self[line_strip[0]] = int(line_strip[1])



"""
DB2 commands
============

Shared parsers for parsing output of the commands of IBM DB2

Db2ls - command ``db2ls -a -c``
-------------------------------
"""
from insights import parser, CommandParser
from insights.parsers import SkipException
from insights.specs import Specs


@parser(Specs.db2ls_a_c)
class Db2ls(CommandParser, list):
    """
    Parse the output of `db2ls -a -c` into a list of dictionaries.

    Typical output of this command::

        #PATH:VRMF:FIXPACK:SPECIAL:INSTALLTIME:INSTALLERUID
        /opt/ibm/db2/V11.5:11.5.6.0:0 ::Fri Jan 14 20:20:07 2022 CST :0
        /opt/ibm/db2/V11.5_01:11.5.7.0:0 ::Fri Feb 11 10:34:51 2022 CST :0

    Examples:
        >>> type(db2ls)
        <class 'insights.parsers.db2.Db2ls'>
        >>> len(db2ls)
        2
        >>> db2ls[0]['PATH']
        '/opt/ibm/db2/V11.5'
        >>> db2ls[1]['PATH']
        '/opt/ibm/db2/V11.5_01'
        >>> db2ls[1]['VRMF']
        '11.5.7.0'
        >>> db2ls[1]['INSTALLERUID']
        '0'
    """

    def parse_content(self, content):
        if not content:
            raise SkipException("Empty.")
        keys = []
        for line in content:
            if not keys and line.startswith('#PATH'):
                keys = [i.strip('#') for i in line.split(':')]
                continue
            line_splits = [i.strip() for i in line.split(':', 4)]
            line_splits.extend([i.strip(': ') for i in line_splits.pop(-1).rsplit(':', 1)])
            self.append(dict(zip(keys, line_splits)))
        if len(self) == 0:
            raise SkipException('Nothing to parse.')



"""
CpuSetCpus - File ``/sys/fs/cgroup/cpuset/cpuset.cpus``
=======================================================

This parser reads the content of ``/sys/fs/cgroup/cpuset/cpuset.cpus``.
This file shows the default cgroup cpuset.cpu of system. The format
of the content is string including comma.
"""

from .. import parser, CommandParser
from insights.specs import Specs


@parser(Specs.cpuset_cpus)
class CpusetCpus(CommandParser):
    """
    Class ``CpusetCpus`` parses the content of the ``/sys/fs/cgroup/cpuset/cpuset.cpus``.

    Attributes:
        cpu_set (list): It is used to show the list of allowed cpu.

        cpu_number (int): It is used to display the number of allowed cpu.

    A small sample of the content of this file looks like::

        0,2-4,7

    Examples:
        >>> type(cpusetinfo)
        <class 'insights.parsers.cpuset_cpus.CpusetCpus'>
        >>> cpusetinfo.cpuset
        ["0", "2", "3", "4", "7"]
        >>> cpusetinfo.cpu_number
        5
    """

    def parse_content(self, content):
        self.cpu_set = []
        self.cpu_number = 0
        values = content[0].strip().split(",")
        for value in values:
            if "-" in value:
                # Parse the value like "2-4"
                start, end = value.split("-")
                self.cpu_set.extend([str(i) for i in range(int(start), int(end) + 1)])
            else:
                self.cpu_set.append(value)
        self.cpu_number = len(self.cpu_set)



from insights.parsers.neutron_server_log import NeutronServerLog
from insights.tests import context_wrap

NEUTRON_LOG = """
2016-09-13 05:56:45.155 30586 WARNING keystonemiddleware.auth_token [-] Identity response: {"error": {"message": "Could not find token: b45405915eb44e608885f894028d37b9", "code": 404, "title": "Not Found"}}
2016-09-13 05:56:45.156 30586 WARNING keystonemiddleware.auth_token [-] Authorization failed for token
2016-09-13 06:06:45.884 30588 WARNING keystonemiddleware.auth_token [-] Authorization failed for token
2016-09-13 06:06:45.886 30588 WARNING keystonemiddleware.auth_token [-] Identity response: {"error": {"message": "Could not find token: fd482ef0ba1144bf944a0a6c2badcdf8", "code": 404, "title": "Not Found"}}
2016-09-13 06:06:45.887 30588 WARNING keystonemiddleware.auth_token [-] Authorization failed for token
2016-09-13 06:06:46.131 30586 WARNING keystonemiddleware.auth_token [-] Authorization failed for token
2016-09-13 06:06:46.131 30586 WARNING keystonemiddleware.auth_token [-] Identity response: {"error": {"message": "Could not find token: bc029dbe33f84fbcb67ef7d592458e60", "code": 404, "title": "Not Found"}}
2016-09-13 06:06:46.132 30586 WARNING keystonemiddleware.auth_token [-] Authorization failed for token
""".strip()


def test_server_log():
    neutron_server = NeutronServerLog(context_wrap(NEUTRON_LOG))
    assert len(neutron_server.get(["WARNING", "Authorization failed for token"])) == 5
    assert len(neutron_server.get(["Identity response:"])) == 3
    assert len(neutron_server.get("Identity response:")) == 3



"""
SSSD_Config - file ``/etc/sssd/sssd.config``
============================================
"""
from insights.core import IniConfigFile
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.sssd_config)
class SSSD_Config(IniConfigFile):
    """
    Parse the content of the ``/etc/sssd/sssd.config`` file.

    The 'sssd' section must always exist.  Within that, the 'domains'
    parameter is usually defined to give a comma-separated list of the
    domains that sssd is to manage.

    The 'sssd' section will define one or more active domains, which are then
    configured in the 'domain/{domain}' section of the configuration.  These
    domains are then available via the 'domains' method, and the configuration
    of a domain can be fetched as a dictionary using the 'domain_config' method.

    Sample configuration::

        [sssd]
        config_file_version = 2

        # Number of times services should attempt to reconnect in the
        # event of a crash or restart before they give up
        reconnection_retries = 3

        # If a back end is particularly slow you can raise this timeout here
        sbus_timeout = 30
        services = nss, pam

        # SSSD will not start if you do not configure any domains.
        # Add new domain configurations as [domain/<NAME>] sections, and
        # then add the list of domains (in the order you want them to be
        # queried) to the "domains" attribute below and uncomment it.
        # domains = LOCAL,LDAP
        domains = example.com
        debug_level = 9

        [nss]
        # The following prevents SSSD from searching for the root user/group in
        # all domains (you can add here a comma-separated list of system accounts that
        # are always going to be /etc/passwd users, or that you want to filter out).
        filter_groups = root
        filter_users = root
        reconnection_retries = 3

        [pam]
        reconnection_retries = 3

        [domain/example.com]
        id_provider = ldap
        lookup_family_order = ipv4_only
        ldap_uri = ldap://ldap.example.com/
        ldap_search_base = dc=example,dc=com
        enumerate = False
        hbase_directory= /home
        create_homedir = True
        override_homedir = /home/%u
        auth_provider = krb5
        krb5_server = kerberos.example.com
        krb5_realm = EXAMPLE.COM

    Example:
        >>> type(conf)
        <class 'insights.parsers.sssd_conf.SSSD_Config'>
        >>> conf.get('nss', 'filter_users')
        'root'
        >>> conf.getint('pam', 'reconnection_retries')
        3
        >>> conf.domains
        ['example.com']
        >>> domain = conf.domain_config('example.com')
        >>> 'ldap_uri' in domain
        True
    """
    @property
    def domains(self):
        """
        Returns the list of domains defined in the 'sssd' section.  This is
        used to refer to the domain-specific sections of the configuration.
        """
        if self.has_option('sssd', 'domains'):
            domains = self.get('sssd', 'domains')
            if domains:
                return domains.split(',')

        # Return a blank list if no domains.
        return []

    def domain_config(self, domain):
        """
        Return the configuration dictionary for a specific domain, given as
        the raw name as listed in the 'domains' property of the sssd section.
        This then looks for the equivalent 'domain/{domain}' section of the
        config file.
        """
        full_domain = 'domain/' + domain
        if full_domain not in self:
            return {}
        return self.items(full_domain)



"""
ResolvConf - file ``/etc/resolv.conf``
======================================

"""

from .. import Parser, parser, get_active_lines, LegacyItemAccess
from insights.specs import Specs

KEYWORD1 = 'search'
KEYWORD2 = 'domain'


@parser(Specs.resolv_conf)
class ResolvConf(LegacyItemAccess, Parser):
    '''
    Parse the ``/etc/resolv.conf`` file into a dictionary of keywords and
    their values.  This is made available via the ``data`` property but the
    object itself can be used as a dictionary thanks to the
    :py:class:`insights.core.LegacyItemAccess` mixin class.

    Each keyword found in the file is stored as a key in the data dictionary,
    storing the list of values given (in order) on all occurrences of that
    keyword in the file.

    According to the man page, the 'domain' and 'search' keywords are
    mutually exclusive. If more than one instance of these keywords is
    present, whichever is last becomes the active resolution method. So, the
    ``active`` key stores which of these keywords was the last present in the
    file.

    Sample file content::

        ; generated by /sbin/dhclient-script
        # This file is being maintained by Puppet.
        # DO NOT EDIT
        search a.b.com b.c.com
        options timeout:2 attempts:2
        nameserver 10.160.224.51
        nameserver 10.61.193.11

    Examples:
        >>> resolv = shared[ResolvConf]
        >>> resolv['active']
        'search'
        >>> resolv['nameserver']
        ["10.160.224.51", "10.61.193.11" ]
        >>> resolv['search']
        ["a.b.com", "b.c.com"]
        >>> resolv.data["options"]  # old style access
        ["timeout:2", "attempts:2"]
    '''

    def parse_content(self, content):

        resolv_info = {}
        name_info = []
        key1_search = key2_domain = False

        # According to the man page, the 'domain' and 'search' keywords are
        # mutually exclusive. If more than one instance of these keywords is
        # present, the last instance wins. So, add a key "active" into
        # resolve_info pointing out which keywords is effective.
        for line in get_active_lines(content):
            # ignore the lines commented by ';'
            if line.startswith(';'):
                continue
            if line.startswith('nameserver'):
                name_info.append(line.split()[1])
            else:
                temp = line.split()
                if temp[0] in (KEYWORD1, KEYWORD2):
                    key1_search = temp[0] == KEYWORD1
                    key2_domain = temp[0] == KEYWORD2
                resolv_info[temp[0]] = temp[1:]

        resolv_info['nameserver'] = name_info
        resolv_info['active'] = ''
        if key1_search and not key2_domain:
            resolv_info['active'] = KEYWORD1
        elif key2_domain and not key1_search:
            resolv_info['active'] = KEYWORD2

        self.data = resolv_info



from insights.tests import context_wrap
from insights.parsers.katello_service_status import KatelloServiceStatus

KS_OUT = """
pulp_celerybeat (pid 443) is running.
celery init v10.0.
Using config script: /etc/default/pulp_workers
node reserved_resource_worker-0 (pid 902) is running...
node reserved_resource_worker-1 (pid 921) is running...
node reserved_resource_worker-2 (pid 938) is running...
node reserved_resource_worker-3 (pid 959) is running...
celery init v10.0.
Using config script: /etc/default/pulp_resource_manager
node resource_manager (pid 691) is running...
tomcat6 is stopped                                         [  OK  ]
dynflow_executor is running.
dynflow_executor_monitor is running.
httpd is stopped
Some services failed to status: tomcat6,httpd
"""

KS_OUT_1 = """
Some services failed to status: httpd
"""

KS_OUT_2 = """
Using config script: /etc/default/pulp_resource_manager
node resource_manager (pid 691) is running...
httpd (pid  16006) is running..
tomcat6 (pid 15560) is running...                          [  OK  ]
Success!
"""


def test_kss():
    kss = KatelloServiceStatus(context_wrap(KS_OUT))
    assert kss.failed_services == ['tomcat6', 'httpd']
    assert not kss.is_ok
    kss = KatelloServiceStatus(context_wrap(KS_OUT_1))
    assert kss.failed_services == ['httpd']
    assert not kss.is_ok
    kss = KatelloServiceStatus(context_wrap(KS_OUT_2))
    assert kss.failed_services == []
    assert kss.is_ok



"""
KeystoneLog - file ``/var/log/keystone/keystone.log``
=====================================================

Module for parsing the log file for Keystone

Typical content of ``keystone.log`` file is::

    2016-11-09 14:31:46.834 818 INFO migrate.versioning.api [-] done
    2016-11-09 14:31:46.834 818 INFO migrate.versioning.api [-] 3 -> 4...
    2016-11-09 14:31:46.872 818 INFO migrate.versioning.api [-] done
    2016-11-09 14:31:48.435 1082 WARNING keystone.assignment.core [-] Deprecated: Use of the identity driver config to automatically configure the same assignment driver has been deprecated, in the "O" release, the assignment driver will need to be expicitly configured if different than the default (SQL).
    2016-11-09 14:31:48.648 1082 WARNING oslo_config.cfg [-] Option "rpc_backend" from group "DEFAULT" is deprecated for removal.  Its value may be silently ignored in the future.
    2016-11-09 14:31:48.680 1082 WARNING oslo_config.cfg [-] Option "rabbit_hosts" from group "oslo_messaging_rabbit" is deprecated for removal.  Its value may be silently ignored in the future.
    2016-11-09 14:31:48.681 1082 WARNING oslo_config.cfg [-] Option "rabbit_userid" from group "oslo_messaging_rabbit" is deprecated for removal.  Its value may be silently ignored in the future.
    2016-11-09 14:31:48.681 1082 WARNING oslo_config.cfg [-] Option "rabbit_password" from group "oslo_messaging_rabbit" is deprecated for removal.  Its value may be silently ignored in the future.
    2016-11-09 14:31:48.774 1082 INFO keystone.cmd.cli [-] Created domain default
    2016-11-09 14:31:48.802 1082 INFO keystone.cmd.cli [req-ace08b7c-d0d2-4b18-b792-1ec3402575b1 - - - - -] Created project admin
    2016-11-09 14:31:48.886 1082 INFO keystone.cmd.cli [req-ace08b7c-d0d2-4b18-b792-1ec3402575b1 - - - - -] Created user admin
    2016-11-09 14:31:48.895 1082 INFO keystone.cmd.cli [req-ace08b7c-d0d2-4b18-b792-1ec3402575b1 - - - - -] Created role admin
    2016-11-09 14:31:48.916 1082 INFO keystone.cmd.cli [req-ace08b7c-d0d2-4b18-b792-1ec3402575b1 - - - - -] Granted admin on admin to user admin.
    2016-11-09 14:31:48.986 1082 INFO keystone.assignment.core [req-ace08b7c-d0d2-4b18-b792-1ec3402575b1 - - - - -] Creating the default role 9fe2ff9ee4384b1894a90878d3e92bab because it does not exist.
    2016-11-09 14:32:09.175 1988 WARNING keystone.assignment.core [-] Deprecated: Use of the identity driver config to automatically configure the same assignment driver has been deprecated, in the "O" release, the assignment driver will need to be expicitly configured if different than the default (SQL).
"""

from .. import LogFileOutput, parser
from insights.specs import Specs


@parser(Specs.keystone_log)
class KeystoneLog(LogFileOutput):
    """Class for parsing ``/var/log/keystone/keystone.log`` file.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`
    """
    pass



import pytest
import doctest
from insights.parsers.user_group import GroupInfo
from insights.parsers import user_group, SkipException, ParseException
from insights.tests import context_wrap

GRP = """
wheel:x:10:admin,tester
mem:x:8:
""".strip()
GRP_NG = """
unknow_case
""".strip()
GRP_EMPTY = ""


def test_grp():
    grp = GroupInfo(context_wrap(GRP))
    assert grp[0]['id'] == 10
    assert grp[0]['name'] == 'wheel'
    assert grp[0]['users'] == ['admin', 'tester']
    assert grp[1]['id'] == 8
    assert grp[1]['name'] == 'mem'
    assert grp[1]['users'] == []
    assert grp.search(users__contains='tester')[0] == grp[0]


def test_ab():
    with pytest.raises(SkipException):
        GroupInfo(context_wrap(GRP_EMPTY))

    with pytest.raises(ParseException):
        GroupInfo(context_wrap(GRP_NG))


def test_doc_examples():
    env = {
        'grp': GroupInfo(context_wrap(GRP))
    }
    failed, total = doctest.testmod(user_group, globs=env)
    assert failed == 0



"""
AWSInstanceType
===============

This parser simply reads the output of command
``curl -s http://169.254.169.254/latest/meta-data/instance-type``,
which is used to check the type of the AWS instance on the host.

"""

from insights.parsers import SkipException, ParseException
from insights import parser, CommandParser
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.aws_instance_type)
class AWSInstanceType(CommandParser):
    """
    .. note::
        This parser is deprecated, please use
        :py:class:`insights.parsers.aws_instance_id.AWSInstanceIdDoc` instead.

    Class for parsing the AWS Instance type returned by command
    ``curl -s http://169.254.169.254/latest/meta-data/instance-type``

    Typical output of this command is::

        r3.xlarge

    Raises:
        SkipException: When content is empty or no parse-able content.
        ParseException: When type cannot be recognized.

    Attributes:
        type (str): The name of AWS instance type in all uppercase letters. E.g. R3, R4, R5, or X1.
        raw (str): The fully type string returned by the ``curl`` command.

    Examples:
        >>> aws_inst.type
        'R3'
        >>> aws_inst.raw
        'r3.xlarge'
    """

    def __init__(self, *args, **kwargs):
        deprecated(AWSInstanceType, "Use AWSInstanceIdDoc in insights.insights.aws_instance_id instead.", "3.0.300")
        super(AWSInstanceType, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        if not content or 'curl: ' in content[0]:
            raise SkipException()

        self.raw = self.type = None

        # Ignore any curl stats that may be present in data
        for l in content:
            if ' ' not in l.strip() and len(l.strip()) > 0:
                self.raw = l.strip()
                if '.' in self.raw:
                    self.type = self.raw.split('.')[0].upper()

        if not self.type:
            raise ParseException('Unrecognized type: "{0}"'.format(content))

    def __repr__(self):
        return "<aws_type: {t}, raw: {r}>".format(t=self.type, r=self.raw)



"""
Utility functions
"""
from __future__ import absolute_import
import glob
import socket
import os
import logging
import uuid
import datetime
import shlex
import re
import sys
import threading
import time
import json
import tarfile
from subprocess import Popen, PIPE, STDOUT

import yaml
try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper

from .. import package_info
from .constants import InsightsConstants as constants
from .collection_rules import InsightsUploadConf, load_yaml

from insights.core.context import Context
from insights.parsers.os_release import OsRelease
from insights.parsers.redhat_release import RedhatRelease

try:
    from insights_client.constants import InsightsConstants as wrapper_constants
except ImportError:
    wrapper_constants = None

logger = logging.getLogger(__name__)


def determine_hostname(display_name=None):
    """
    Find fqdn if we can
    """
    if display_name:
        # if display_name is provided, just return the given name
        return display_name
    else:
        socket_gethostname = socket.gethostname()
        socket_fqdn = socket.getfqdn()

        try:
            socket_ex = socket.gethostbyname_ex(socket_gethostname)[0]
        except (LookupError, socket.gaierror):
            socket_ex = ''

        gethostname_len = len(socket_gethostname)
        fqdn_len = len(socket_fqdn)
        ex_len = len(socket_ex)

        if fqdn_len > gethostname_len or ex_len > gethostname_len:
            if "localhost" not in socket_ex and len(socket_ex):
                return socket_ex
            if "localhost" not in socket_fqdn:
                return socket_fqdn

        return socket_gethostname


def get_time():
    return datetime.datetime.isoformat(datetime.datetime.now())


def write_registered_file():
    delete_unregistered_file()
    for f in constants.registered_files:
        if os.path.lexists(f):
            if os.path.islink(f):
                # kill symlinks and regenerate
                os.remove(f)
                write_to_disk(f)
        else:
            write_to_disk(f)


def write_unregistered_file(date=None):
    """
    Write .unregistered out to disk
    """
    delete_registered_file()
    if date is None:
        date = get_time()
    for f in constants.unregistered_files:
        if os.path.lexists(f):
            if os.path.islink(f):
                # kill symlinks and regenerate
                os.remove(f)
                write_to_disk(f, content=str(date))
        else:
            write_to_disk(f, content=str(date))


def delete_registered_file():
    for f in constants.registered_files:
        write_to_disk(f, delete=True)


def delete_unregistered_file():
    for f in constants.unregistered_files:
        write_to_disk(f, delete=True)


def delete_cache_files():
    for f in glob.glob(os.path.join(constants.insights_core_lib_dir, "*.json")):
        os.remove(f)


def write_to_disk(filename, delete=False, content=get_time()):
    """
    Write filename out to disk
    """
    if not os.path.exists(os.path.dirname(filename)):
        return
    if delete:
        if os.path.lexists(filename):
            os.remove(filename)
    else:
        with open(filename, 'wb') as f:
            f.write(content.encode('utf-8'))


def generate_machine_id(new=False,
                        destination_file=constants.machine_id_file):
    """
    Generate a machine-id if /etc/insights-client/machine-id does not exist
    """
    machine_id = None
    machine_id_file = None
    logging_name = 'machine-id'

    if os.path.isfile(destination_file) and not new:
        logger.debug('Found %s', destination_file)
        with open(destination_file, 'r') as machine_id_file:
            machine_id = machine_id_file.read()
    else:
        logger.debug('Could not find %s file, creating', logging_name)
        machine_id = str(uuid.uuid4())
        logger.debug("Creating %s", destination_file)
        write_to_disk(destination_file, content=machine_id)

    try:
        return str(uuid.UUID(str(machine_id).strip(), version=4))
    except ValueError as e:
        logger.error("Invalid machine ID: %s", machine_id)
        logger.error("Error details: %s", str(e))
        logger.error("Remove %s and a new one will be generated.\nRerun the client with --register", destination_file)
        sys.exit(constants.sig_kill_bad)


def _expand_paths(path):
    """
    Expand wildcarded paths
    """
    dir_name = os.path.dirname(path)
    paths = []
    logger.debug("Attempting to expand %s", path)
    if os.path.isdir(dir_name):
        files = os.listdir(dir_name)
        match = os.path.basename(path)
        for file_path in files:
            if re.match(match, file_path):
                expanded_path = os.path.join(dir_name, file_path)
                paths.append(expanded_path)
        logger.debug("Expanded paths %s", paths)
        return paths
    else:
        logger.debug("Could not expand %s", path)


def validate_remove_file(config):
    """
    Validate the remove file and tags file
    """
    return InsightsUploadConf(config).validate()


def write_data_to_file(data, filepath):
    '''
    Write data to file
    '''
    try:
        os.makedirs(os.path.dirname(filepath), 0o700)
    except OSError:
        pass

    write_to_disk(filepath, content=data)


def magic_plan_b(filename):
    '''
    Use this in instances where
    python-magic is MIA and can't be installed
    for whatever reason
    '''
    cmd = shlex.split('file --mime-type --mime-encoding ' + filename)
    stdout, stderr = Popen(cmd, stdout=PIPE).communicate()
    stdout = stdout.decode("utf-8")
    mime_str = stdout.split(filename + ': ')[1].strip()
    return mime_str


def run_command_get_output(cmd):
    proc = Popen(shlex.split(cmd),
                 stdout=PIPE, stderr=STDOUT)
    stdout, stderr = proc.communicate()

    return {
        'status': proc.returncode,
        'output': stdout.decode('utf-8', 'ignore')
    }


def modify_config_file(updates):
    '''
    Update the config file with certain things
    '''
    cmd = '/bin/sed '
    for key in updates:
        cmd = cmd + '-e \'s/^#*{key}.*=.*$/{key}={value}/\' '.format(key=key, value=updates[key])
    cmd = cmd + constants.default_conf_file
    status = run_command_get_output(cmd)
    write_to_disk(constants.default_conf_file, content=status['output'])


def get_version_info():
    '''
    Get the insights client and core versions for archival
    '''
    try:
        client_version = wrapper_constants.version
    except AttributeError:
        # wrapper_constants is None or has no attribute "version"
        client_version = None
    version_info = {}
    version_info['core_version'] = '%s-%s' % (package_info['VERSION'], package_info['RELEASE'])
    version_info['client_version'] = client_version
    return version_info


def print_egg_versions():
    '''
    Log all available eggs' versions
    '''
    versions = get_version_info()
    logger.debug('Client version: %s', versions['client_version'])
    logger.debug('Core version: %s', versions['core_version'])
    logger.debug('All egg versions:')
    eggs = [
        os.getenv('EGG'),
        '/var/lib/insights/newest.egg',
        '/var/lib/insights/last_stable.egg',
        '/etc/insights-client/rpm.egg',
    ]
    if not sys.executable:
        logger.debug('Python executable not found.')
        return

    for egg in eggs:
        if egg is None:
            logger.debug('ENV egg not defined.')
            continue
        if not os.path.exists(egg):
            logger.debug('%s not found.', egg)
            continue
        try:
            proc = Popen([sys.executable, '-c',
                         'from insights import package_info; print(\'%s-%s\' % (package_info[\'VERSION\'], package_info[\'RELEASE\']))'],
                         env={'PYTHONPATH': egg, 'PATH': os.getenv('PATH')}, stdout=PIPE, stderr=STDOUT)
        except OSError:
            logger.debug('Could not start python.')
            return
        stdout, stderr = proc.communicate()
        version = stdout.decode('utf-8', 'ignore').strip()
        logger.debug('%s: %s', egg, version)


def read_pidfile():
    '''
    Read the pidfile we wrote at launch
    '''
    pid = None
    try:
        with open(constants.pidfile) as pidfile:
            pid = pidfile.read()
    except IOError:
        logger.debug('Could not open pidfile for reading.')
    return pid


def _systemd_notify(pid):
    '''
    Ping the systemd watchdog with the main PID so that
    the watchdog doesn't kill the process
    '''
    try:
        proc = Popen(['/usr/bin/systemd-notify', '--pid=' + str(pid), 'WATCHDOG=1'])
    except OSError as e:
        logger.debug('Could not launch systemd-notify: %s', str(e))
        return False
    stdout, stderr = proc.communicate()
    if proc.returncode != 0:
        logger.debug('systemd-notify returned %s', proc.returncode)
        return False
    return True


def systemd_notify_init_thread():
    '''
    Use a thread to periodically ping systemd instead
    of calling it on a per-command basis
    '''
    pid = read_pidfile()
    if not pid:
        logger.debug('No PID specified.')
        return
    if not os.getenv('NOTIFY_SOCKET'):
        # running standalone, not via systemd job
        return
    if not os.path.exists('/usr/bin/systemd-notify'):
        # RHEL 6, no systemd
        return

    def _sdnotify_loop():
        while True:
            # run sdnotify every 30 seconds
            if not _systemd_notify(pid):
                # end the loop if something goes wrong
                break
            time.sleep(30)

    sdnotify_thread = threading.Thread(target=_sdnotify_loop, args=())
    sdnotify_thread.daemon = True
    sdnotify_thread.start()


def get_tags(tags_file_path=constants.default_tags_file):
    '''
    Load tag data from the tags file.

    Returns: a dict containing tags defined on the host.
    '''
    tags = None

    if os.path.isfile(tags_file_path):
        try:
            tags = load_yaml(tags_file_path)
        except RuntimeError:
            logger.error("Invalid YAML. Unable to load %s", tags_file_path)
            return None
    else:
        logger.debug("%s does not exist", tags_file_path)

    return tags


def write_tags(tags, tags_file_path=constants.default_tags_file):
    """
    Writes tags to tags_file_path

    Arguments:
      - tags (dict): the tags to write
      - tags_file_path (string): path to which tag data will be written

    Returns: None
    """
    with open(tags_file_path, mode="w") as f:
        data = yaml.dump(tags, Dumper=Dumper, default_flow_style=False)
        f.write(data)


def migrate_tags():
    '''
    We initially released the tags feature with the tags file set as
    tags.conf, but soon after switched it over to tags.yaml. There may be
    installations out there with tags.conf files, so rename the files.
    '''
    tags_conf = os.path.join(constants.default_conf_dir, 'tags.conf')
    tags_yaml = os.path.join(constants.default_conf_dir, 'tags.yaml')

    if os.path.exists(tags_yaml):
        # current default file exists, do nothing
        return
    if os.path.exists(tags_conf):
        # old file exists and current does not
        logger.info('Tags file %s detected. This filename is deprecated; please use %s. The file will be renamed automatically.',
                    tags_conf, tags_yaml)
        try:
            os.rename(tags_conf, tags_yaml)
        except OSError as e:
            logger.error(e)


def get_parent_process():
    '''
    Get parent process of the client

    Returns: string
    '''
    ppid = os.getppid()
    output = run_command_get_output('cat /proc/%s/status' % ppid)
    if output['status'] == 0:
        name = output['output'].splitlines()[0].split('\t')[1]
        return name
    else:
        return "unknown"


def os_release_info():
    '''
    Use insights-core to fetch the os-release or redhat-release info

    Returns a tuple of OS name and version
    '''
    os_family = "Unknown"
    os_release = ""
    for p in ["/etc/os-release", "/etc/redhat-release"]:
        try:
            with open(p) as f:
                data = f.readlines()

            ctx = Context(content=data, path=p, relative_path=p)
            if p == "/etc/os-release":
                rls = OsRelease(ctx)
                os_family = rls.data.get("NAME")
                os_release = rls.data.get("VERSION_ID")
            elif p == "/etc/redhat-release":
                rls = RedhatRelease(ctx)
                os_family = rls.product
                os_release = rls.version
            break
        except IOError:
            continue
        except Exception as e:
            logger.warning("Failed to detect OS version: %s", e)
    return (os_family, os_release)


def largest_spec_in_archive(archive_file):
    logger.info("Checking for large files...")
    tar_file = tarfile.open(archive_file, 'r')
    largest_fsize = 0
    largest_file_name = ""
    largest_spec = ""
    # get the name of the archive
    name = os.path.basename(archive_file).split(".tar.gz")[0]
    # get the archives from inside meta_data directory
    metadata_top = os.path.join(name, "meta_data/")
    data_top = os.path.join(name, "data")
    for file in tar_file.getmembers():
        if metadata_top in file.name:
            file_extract = tar_file.extractfile(file.name)
            specs_metadata = json.load(file_extract)
            results = specs_metadata.get("results", [])
            if not results:
                continue
            if not isinstance(results, list):
                # specs with only one resulting file are not in list form
                results = [results]
            for result in results:
                # get the path of the spec result and check its filesize
                fname = result.get("object", {}).get("relative_path")
                abs_fname = os.path.join('.', data_top, fname)
                # get the archives from inside data directory
                data_file = tar_file.getmember(abs_fname)
                if (data_file.size > largest_fsize):
                    largest_fsize = data_file.size
                    largest_file_name = fname
                    largest_spec = specs_metadata["name"]
    return (largest_file_name, largest_fsize, largest_spec)


def size_in_mb(num_bytes):
    return float(num_bytes) / (1024 * 1024)



"""
multipath.conf file content
===========================

The base class is the MultipathConfParser class, which reads the multipath
daemon's ``/etc/multipath.conf`` configuration file.
This is in a pseudo-JSON format.

MultipathConf - file ``/etc/multipath.conf``
--------------------------------------------

MultipathConfInitramfs - command ``lsinitrd -f /etc/multipath.conf``
--------------------------------------------------------------------

"""

import string
from insights.contrib import pyparsing as p
from insights import parser, Parser, LegacyItemAccess
from insights.core import ConfigParser
from insights.parsers import SkipException
from insights.parsr import (EOF, Forward, LeftCurly, Lift, Literal, LineEnd,
        RightCurly, Many, Number, OneLineComment, PosMarker, skip_none, String,
        QuotedString, WS, WSChar)
from insights.parsr.query import Entry
from insights.specs import Specs


class MultipathConfParser(Parser, LegacyItemAccess):
    """
    Shared parser for the file ``/etc/multipath.conf`` and output of
    ``lsinitrd -f /etc/multipath.conf`` applied to
    /boot/initramfs-<kernel-version>.img.

    Return a dict where the keys are the name of sections in multipath
    configuraion file.  If there are subsections, the value is a list of
    dictionaries with parameters as key and value.  Otherwise the value is
    just a single dictionary.

    Configuration File Example::

        defaults {
               path_selector           "round-robin 0"
               user_friendly_names      yes
        }

        multipaths {
               multipath {
                       alias                   yellow
                       path_grouping_policy    multibus
              }
               multipath {
                       wwid                    1DEC_____321816758474
                       alias                   red
              }
        }

        devices {
               device {
                       path_selector           "round-robin 0"
                       no_path_retry            queue
              }
               device {
                       vendor                  1DEC_____321816758474
                       path_grouping_policy    red
              }
        }

        blacklist {
              wwid 26353900f02796769
              devnode "^hd[a-z]"
        }

    Parse Result::

        data = {
          "blacklist": {
            "devnode": "^hd[a-z]",
            "wwid": "26353900f02796769"
          },
          "devices": [
            {
              "path_selector": "round-robin 0",
              "no_path_retry": "queue"
            },
            {
              "path_grouping_policy": "red",
              "vendor": "1DEC_____321816758474"
            }
          ],
          "defaults": {
            "path_selector": "round-robin 0",
            "user_friendly_names": "yes"
          },
          "multipaths": [
            {
              "alias": "yellow",
              "path_grouping_policy": "multibus"
            },
            {
              "alias": "red",
              "wwid": "1DEC_____321816758474"
            }
          ]
        }
    """

    @classmethod
    def _create_parser(cls):
        """
        Need to check the specific symbol "/" in attr_value part as well.
        I checked some multipath configuraion files from the sosreport and got
        although there are some more specific symbols like "-%", it is enclosed
        in double quotes and will be accepted. Furthermore, I also checked the
        source code of "device-mapper-multipath" and got if the attr_value in
        "multipath.conf" include a "whitespace", it must be enclosed in double
        quotation marks. So, we could just add one more specific symbol "/" to
        check.
        ----------------------------------------------------------
        udev_dir                /dev
        getuid_callout          "/sbin/scsi_id -g -u -s /block/%n"
        ----------------------------------------------------------
        """
        section_name = p.Word(p.alphas + "_")
        attr_name = attr_value = p.Word(p.alphanums + "_/")
        LBRACE, RBRACE = map(p.Suppress, "{}")
        attr = p.Group(attr_name + (attr_value | p.quotedString.setParseAction(p.removeQuotes)))
        attr_list = p.Dict(p.ZeroOrMore(attr))
        simple_section = p.Group(section_name + LBRACE + attr_list + RBRACE)
        complex_section = p.Group(section_name + LBRACE + p.OneOrMore(simple_section) + RBRACE)
        simple_or_complex = p.Dict(simple_section | complex_section)
        my_conf = p.Group(p.ZeroOrMore(simple_or_complex))
        my_conf.ignore("#" + p.restOfLine)
        return my_conf

    def parse_content(self, content):
        if not content:
            raise SkipException("Empty content.")
        self.data = MultipathConfParser._create_parser().parseString("\n".join(content))[0].asDict()


@parser(Specs.multipath_conf)
class MultipathConf(MultipathConfParser):
    """
    Parser for the file ``/etc/multipath.conf``.

    Examples:
        >>> conf = shared[MultipathConf]
        >>> conf.data['blacklist']['devnode']  # Access via data property
        '^hd[a-z]'
        >>> conf['defaults']['user_friendly_names']  # Pseudo-dict access
        'yes'
        >>> len(conf['multipaths'])
        2
        >>> conf['multipaths'][0]['alias']
        'yellow'
    """
    pass


@parser(Specs.multipath_conf_initramfs)
class MultipathConfInitramfs(MultipathConfParser):
    """
    Parser for the output of ``lsinitrd -f /etc/multipath.conf`` applied to
    /boot/initramfs-<kernel-version>.img.

    Examples:
        >>> conf = shared[MultipathConfInitramfs]
        >>> conf.data['blacklist']['devnode']  # Access via data property
        '^hd[a-z]'
        >>> conf['defaults']['user_friendly_names']  # Pseudo-dict access
        'yes'
        >>> len(conf['multipaths'])
        2
        >>> conf['multipaths'][0]['alias']
        'yellow'
    """
    pass


def parse_doc(content, ctx):
    def to_entry(name, rest):
        if isinstance(rest, list):
            return Entry(name=name.value, children=rest, lineno=name.lineno, src=ctx)
        return Entry(name=name.value, attrs=[rest], lineno=name.lineno, src=ctx)

    Stmt = Forward()
    Num = Number & (WSChar | LineEnd)
    NULL = Literal("none", value=None)
    Comment = (WS >> OneLineComment("#").map(lambda x: None))
    BeginBlock = (WS >> LeftCurly << WS)
    EndBlock = (WS >> RightCurly << WS)
    Bare = String(set(string.printable) - (set(string.whitespace) | set("#{}'\"")))
    Name = WS >> PosMarker(String(string.ascii_letters + "_")) << WS
    Value = WS >> (Num | NULL | QuotedString | Bare) << WS
    Block = BeginBlock >> Many(Stmt).map(skip_none) << EndBlock
    Stanza = (Lift(to_entry) * Name * (Block | Value)) | Comment
    Stmt <= WS >> Stanza << WS
    Doc = Many(Stmt).map(skip_none)
    Top = Doc + EOF

    return Entry(children=Top(content)[0])


@parser(Specs.multipath_conf)
class MultipathConfTree(ConfigParser):
    """
    Exposes multipath configuration through the parsr query interface.

    See the :py:class:`insights.core.ConfigComponent` class for example usage.
    """
    def parse_doc(self, content):
        return parse_doc("\n".join(content), ctx=self)


def get_tree(root=None):
    """
    This is a helper function to get a multipath configuration component for
    your local machine or an archive. It's for use in interactive sessions.
    """
    from insights import run
    return run(MultipathConfTree, root=root).get(MultipathConfTree)


@parser(Specs.multipath_conf_initramfs)
class MultipathConfTreeInitramfs(ConfigParser):
    """
    Exposes the multipath configuration from initramfs image through the
    parsr query interface.

    See the :py:class:`insights.core.ConfigComponent` class for example usage.
    """
    def parse_doc(self, content):
        return parse_doc("\n".join(content), ctx=self)


def get_tree_from_initramfs(root=None):
    """
    This is a helper function to get a multipath configuration(from initramfs
    image) component for your local machine or an archive. It's for use in
    interactive sessions.
    """
    from insights import run
    return run(MultipathConfTreeInitramfs, root=root).get(MultipathConfTreeInitramfs)



from insights.parsr.query import all_, any_, compile_queries, Entry, lt

complex_tree = Entry(name="root",
                    attrs=[1, 2, 3, 4],
                    children=[
                        Entry(name="child", attrs=[1, 1, 2]),
                        Entry(name="child", attrs=[1, 1, 2, 3, 5]),
                        Entry(name="child", attrs=[1, 1, 3, 5, 9]),
                        Entry(name="dog", attrs=["woof"], children=[
                            Entry(name="puppy", attrs=["smol"]),
                            Entry(name="puppy", attrs=["fluffy"]),
                            Entry(name="kitten", attrs=["wut"]),
                        ])
                    ])


def test_complex():
    t = complex_tree

    q = compile_queries("child")
    assert len(q(t.children)) == 3

    q = compile_queries(("child", 3))
    assert len(q(t.children)) == 2

    q = compile_queries(("child", all_(lt(3))))
    assert len(q(t.children)) == 1

    q = compile_queries(("child", any_(1)))
    assert len(q(t.children)) == 3

    q = compile_queries("dog", "puppy")
    assert len(q(t.children)) == 2



import doctest
import pytest
from insights.parsers import lsinitrd
from insights.tests import context_wrap


LSINITRD_ALL = """
Image: /boot/initramfs-3.10.0-862.el7.x86_64.img: 24M
========================================================================
Early CPIO image
========================================================================
drwxr-xr-x   3 root     root            0 Apr 20 15:58 .
-rw-r--r--   1 root     root            2 Apr 20 15:58 early_cpio
drwxr-xr-x   3 root     root            0 Apr 20 15:58 kernel
drwxr-xr-x   3 root     root            0 Apr 20 15:58 kernel/x86
drwxr-xr-x   2 root     root            0 Apr 20 15:58 kernel/x86/microcode
-rw-r--r--   1 root     root        12684 Apr 20 15:58 kernel/x86/microcode/AuthenticAMD.bin
========================================================================
Version: dracut-033-535.el7

Arguments: -f

dracut modules:
bash
nss-softokn
i18n
network
ifcfg
drm
plymouth
dm
kernel-modules
lvm
resume
rootfs-block
terminfo
udev-rules
biosdevname
systemd
usrmount
base
fs-lib
shutdown
========================================================================
drwxr-xr-x  12 root     root            0 Apr 20 15:58 .
crw-r--r--   1 root     root       5,   1 Apr 20 15:57 dev/console
crw-r--r--   1 root     root       1,  11 Apr 20 15:57 dev/kmsg
crw-r--r--   1 root     root       1,   3 Apr 20 15:57 dev/null
lrwxrwxrwx   1 root     root            7 Apr 20 15:57 bin -> usr/bin
========================================================================
""".strip()

LSINITRD_FILTERED = """
drwxr-xr-x   3 root     root            0 Apr 20 15:58 kernel/x86
Version: dracut-033-535.el7
dracut modules:
kernel-modules
udev-rules
drwxr-xr-x  12 root     root            0 Apr 20 15:58 .
crw-r--r--   1 root     root       5,   1 Apr 20 15:57 dev/console
crw-r--r--   1 root     root       1,  11 Apr 20 15:57 dev/kmsg
crw-r--r--   1 root     root       1,   3 Apr 20 15:57 dev/null
""".strip()

LSINITRD_EMPTY = ""

# This test case
LSINITRD_BROKEN = """
drwxr-xr-x   3 root     root            0 Apr 20 15:58 kernel/x86
Version: dracut-033-535.el7
dracut modules:
kernel-modules
udev-rules
drwxr-xr-x  12 root     root            0 Apr 20 15:58 .
crw-r--r--   1
crw-r-
c
""".strip()


def test_lsinitrd_empty():
    d = lsinitrd.Lsinitrd(context_wrap(LSINITRD_EMPTY))
    assert len(d.data) == 0
    assert d.search(name__contains='kernel') == []
    assert d.unparsed_lines == []


def test_lsinitrd_filtered():
    d = lsinitrd.Lsinitrd(context_wrap(LSINITRD_FILTERED))
    assert len(d.data) == 5
    assert d.search(name__contains='kernel') == [{'type': 'd', 'perms': 'rwxr-xr-x', 'links': 3, 'owner': 'root', 'group': 'root', 'size': 0, 'date': 'Apr 20 15:58', 'name': 'kernel/x86', 'raw_entry': 'drwxr-xr-x   3 root     root            0 Apr 20 15:58 kernel/x86', 'dir': ''}]
    assert d.unparsed_lines == ['Version: dracut-033-535.el7', 'dracut modules:', 'kernel-modules', 'udev-rules']


def test_lsinitrd_all():
    d = lsinitrd.Lsinitrd(context_wrap(LSINITRD_ALL))
    lsdev = d.search(name__contains='dev')
    assert len(lsdev) == 3
    dev_console = {
        'type': 'c', 'perms': 'rw-r--r--', 'links': 1, 'owner': 'root', 'group': 'root',
        'major': 5, 'minor': 1, 'date': 'Apr 20 15:57', 'name': 'dev/console', 'dir': '',
        'raw_entry': 'crw-r--r--   1 root     root       5,   1 Apr 20 15:57 dev/console'
    }
    assert dev_console in lsdev
    assert 'dev/kmsg' in [l['name'] for l in lsdev]
    assert 'dev/null' in [l['name'] for l in lsdev]
    assert len(d.data) == 10
    assert len(d.unparsed_lines) == 32
    assert "udev-rules" in d.unparsed_lines


def test_lsinitrd_broken():
    """
    For this testcase, ls_parser.parse() will throw an IndexError.
    Assert with this specific error here.
    """
    with pytest.raises(Exception) as err:
        lsinitrd.Lsinitrd(context_wrap(LSINITRD_BROKEN))
    assert "list index out of range" in str(err)


def test_lsinitrd_docs():
    failed_count, tests = doctest.testmod(
        globs={'ls': lsinitrd.Lsinitrd(context_wrap(LSINITRD_FILTERED))}
    )
    assert failed_count == 0



"""
pluginconf.d configuration file - Files
=======================================

Shared mappers for parsing and extracting data from
``/etc/yum/pluginconf.d/*.conf`` files. Parsers contained
in this module are:

PluginConfD - files ``/etc/yum/pluginconf.d/*.conf``
---------------------------------------------------

PluginConfDIni - files ``/etc/yum/pluginconf.d/*.conf``
-------------------------------------------------------
"""
from insights.core import IniConfigFile, LegacyItemAccess, Parser
from insights.core.plugins import parser
from insights.parsers import get_active_lines
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.pluginconf_d)
class PluginConfD(LegacyItemAccess, Parser):
    """
    .. warning::
        This parser is deprecated, please use
        :py:class:`insights.parsers.pluginconf_d.PluginConfDIni` instead

    Class to parse configuration file under ``pluginconf.d``

    Sample configuration::

        [main]
        enabled = 0
        gpgcheck = 1
        timeout = 120

        # You can specify options per channel, e.g.:
        #
        #[rhel-i386-server-5]
        #enabled = 1
        #
        #[some-unsigned-custom-channel]
        #gpgcheck = 0
    """
    def parse_content(self, content):
        deprecated(PluginConfD, "Deprecated. Use 'PluginConfDIni' instead.", "3.0.300")
        plugin_dict = {}
        section_dict = {}
        key = None
        for line in get_active_lines(content):
            if line.startswith('['):
                section_dict = {}
                plugin_dict[line[1:-1]] = section_dict
            elif '=' in line:
                key, _, value = line.partition("=")
                key = key.strip()
                section_dict[key] = value.strip()
            else:
                if key:
                    section_dict[key] = ','.join([section_dict[key], line])
        self.data = plugin_dict

    def __iter__(self):
        for sec in self.data:
            yield sec


@parser(Specs.pluginconf_d)
class PluginConfDIni(IniConfigFile):
    """
    Read yum plugin config files, in INI format, using the standard INI file
    parser class.

    Sample configuration::

        [main]
        enabled = 0
        gpgcheck = 1
        timeout = 120

        # You can specify options per channel, e.g.:
        #
        #[rhel-i386-server-5]
        #enabled = 1
        #
        #[some-unsigned-custom-channel]
        #gpgcheck = 0

        [test]
        test_multiline_config = http://example.com/repos/test/
                                http://mirror_example.com/repos/test/

    Examples:
        >>> type(conf)
        <class 'insights.parsers.pluginconf_d.PluginConfDIni'>
        >>> conf.sections()
        ['main', 'test']
        >>> conf.has_option('main', 'gpgcheck')
        True
        >>> conf.get("main", "enabled")
        '0'
        >>> conf.getint("main", "timeout")
        120
        >>> conf.getboolean("main", "enabled")
        False
        >>> conf.get("test", "test_multiline_config")
        'http://example.com/repos/test/ http://mirror_example.com/repos/test/'
    """
    pass



"""
Hostname
========

Combiner for ``hostname`` information. It uses the results of all the
``Hostname`` parsers and the ``SystemID`` parser to get the fqdn,
hostname and domain information.


"""

from insights.core.plugins import combiner
from insights.core.serde import deserializer, serializer
from insights.parsers.hostname import Hostname as HnF, HostnameShort as HnS, HostnameDefault as HnD
from insights.parsers.systemid import SystemID
from insights.util import deprecated


@combiner([HnF, HnD, HnS, SystemID])
class Hostname(object):
    """
    Check hostname and systemid to get the fqdn, hostname and domain.

    Prefer hostname to systemid.

    Examples:
        >>> type(hostname)
        <class 'insights.combiners.hostname.Hostname'>
        >>> hostname.fqdn
        'rhel7.example.com'
        >>> hostname.hostname
        'rhel7'
        >>> hostname.domain
        'example.com'

    Raises:
        Exception: If no hostname can be found in any of the source parsers.
    """
    def __init__(self, hf, hd, hs, sid):
        self.fqdn = self.hostname = self.domain = None

        if hf or hs or hd:
            hn = hf or hs or hd
            self.hostname = self.fqdn = hn.hostname
            self.domain = ''
            if hf and hf.fqdn:
                self.fqdn = hf.fqdn
                self.domain = hf.domain
        else:
            self.fqdn = sid.get("profile_name")
            if self.fqdn:
                self.hostname = self.fqdn.split(".")[0]
                self.domain = ".".join(self.fqdn.split(".")[1:])

        if not self.hostname or not self.fqdn:
            raise Exception("Unable to get hostname.")


@combiner([HnF, HnD, HnS, SystemID])
def hostname(hf, hd, hs, sid):
    """
    .. warning::
        This combiner methode is deprecated, please use
        :py:class:`insights.combiners.hostname.Hostname` instead.

    Check hostname and systemid to get the fqdn, hostname and domain.

    Prefer hostname to systemid.

    Examples:
        >>> hn.fqdn
        'rhel7.example.com'
        >>> hn.hostname
        'rhel7'
        >>> hn.domain
        'example.com'

    Returns:
        insights.combiners.hostname.Hostname: A class with `fqdn`,
        `hostname` and `domain` attributes.

    Raises:
        Exception: If no hostname can be found in any of the source parsers.
    """
    deprecated(hostname, "Use the `Hostname` class instead.", "3.0.300")
    return Hostname(hf, hd, hs, sid)


@serializer(Hostname)
def serialize(obj, root=None):
    return {"fqdn": obj.fqdn, "hostname": obj.hostname, "domain": obj.domain}


@deserializer(Hostname)
def deserialize(_type, data, root=None):
    return Hostname(**data)



import doctest
import pytest

from insights.parsers import ceph_cmd_json_parsing, ParseException
from insights.parsers.ceph_cmd_json_parsing import CephOsdDump, CephOsdDf, CephS, CephECProfileGet, CephCfgInfo, \
    CephHealthDetail, CephDfDetail, CephOsdTree, CephReport
from insights.tests.parsers import skip_exception_check
from insights.tests import context_wrap

CEPH_OSD_DUMP_INFO = """
{
    "epoch": 210,
    "fsid": "2734f9b5-2013-48c1-8e96-d31423444717",
    "created": "2016-11-12 16:08:46.307206",
    "modified": "2017-03-07 08:55:53.301911",
    "flags": "sortbitwise",
    "cluster_snapshot": "",
    "pool_max": 12,
    "max_osd": 8,
    "pools": [
        {
            "pool": 0,
            "pool_name": "rbd",
            "flags": 1,
            "flags_names": "hashpspool",
            "type": 1,
            "size": 3,
            "min_size": 2,
            "crush_ruleset": 0,
            "object_hash": 2,
            "pg_num": 256
        }
    ]
}
""".strip()

CEPH_OSD_DF_INFO = """
{
    "nodes": [
        {
            "id": 0,
            "name": "osd.0",
            "type": "osd",
            "type_id": 0,
            "crush_weight": 1.091095,
            "depth": 2,
            "reweight": 1.000000,
            "kb": 1171539620,
            "kb_used": 4048208,
            "kb_avail": 1167491412,
            "utilization": 0.345546,
            "var": 1.189094,
            "pgs": 945
        }
    ],
    "stray": [],
    "summary": {
        "total_kb": 8200777340,
        "total_kb_used": 23831128,
        "total_kb_avail": 8176946212,
        "average_utilization": 0.290596,
        "min_var": 0.803396,
        "max_var": 1.189094,
        "dev": 0.035843
    }
}
""".strip()

CEPH_S_INFO = """
{
    "health": {

    },
    "pgmap": {
        "pgs_by_state": [
            {
                "state_name": "active+clean",
                "count": 1800
            }
        ],
        "version": 314179,
        "num_pgs": 1800,
        "data_bytes": 7943926574,
        "bytes_used": 24405610496,
        "bytes_avail": 8373190385664,
        "bytes_total": 8397595996160
    },
    "fsmap": {
        "epoch": 1,
        "by_rank": []
    }
}
""".strip()

CEPH_DF_DETAIL_INFO = """
{
    "stats": {
        "total_bytes": 17113243648,
        "total_used_bytes": 203120640,
        "total_avail_bytes": 16910123008,
        "total_objects": 0
    },
    "pools": [
        {
            "name": "rbd",
            "id": 0,
            "stats": {
                "kb_used": 0,
                "bytes_used": 0,
                "max_avail": 999252180,
                "objects": 0,
                "dirty": 0,
                "rd": 0,
                "rd_bytes": 0,
                "wr": 0,
                "wr_bytes": 0,
                "raw_bytes_used": 0
            }
        },
        {
            "name": "ecpool",
            "id": 2,
            "stats": {
                "kb_used": 0,
                "bytes_used": 0,
                "max_avail": 1998504360,
                "objects": 0,
                "dirty": 0,
                "rd": 0,
                "rd_bytes": 0,
                "wr": 0,
                "wr_bytes": 0,
                "raw_bytes_used": 0
            }
        }
    ]
}
""".strip()

CEPH_HEALTH_DETAIL_INFO = """
{
    "health": {
    },
    "timechecks": {
        "epoch": 4,
        "round": 0,
        "round_status": "finished"
    },
    "summary": [],
    "overall_status": "HEALTH_OK",
    "detail": []
}
""".strip()

CEPH_OSD_EC_PROFILE_GET = """
{
    "k": "2",
    "m": "1",
    "plugin": "jerasure",
    "technique": "reed_sol_van"
}
""".strip()

CEPHINFO = """
{
    "name": "osd.1",
    "cluster": "ceph",
    "debug_none": "0/5",
    "heartbeat_interval": "5",
    "heartbeat_file": "",
    "heartbeat_inject_failure": "0",
    "perf": "true",
    "max_open_files": "131072",
    "ms_type": "simple",
    "ms_tcp_nodelay": "true",
    "ms_tcp_rcvbuf": "0",
    "ms_tcp_prefetch_max_size": "4096",
    "ms_initial_backoff": "0.2",
    "ms_max_backoff": "15",
    "ms_crc_data": "true",
    "ms_crc_header": "true",
    "ms_die_on_bad_msg": "false",
    "ms_die_on_unhandled_msg": "false",
    "ms_die_on_old_message": "false",
    "ms_die_on_skipped_message": "false",
    "ms_dispatch_throttle_bytes": "104857600",
    "ms_bind_ipv6": "false",
    "ms_bind_port_min": "6800",
    "ms_bind_port_max": "7300",
    "ms_bind_retry_count": "3",
    "ms_bind_retry_delay": "5"
}
""".strip()

CEPH_OSD_TREE = """
{
    "nodes": [
        {
            "id": -1,
            "name": "default",
            "type": "root",
            "type_id": 10,
            "children": [
                -5,
                -4,
                -3,
                -2
            ]
        },
        {
            "id": -2,
            "name": "dhcp-192-56",
            "type": "host",
            "type_id": 1,
            "children": []
        },
        {
            "id": -3,
            "name": "dhcp-192-104",
            "type": "host",
            "type_id": 1,
            "children": []
        },
        {
            "id": -4,
            "name": "dhcp-192-67",
            "type": "host",
            "type_id": 1,
            "children": []
        },
        {
            "id": -5,
            "name": "localhost",
            "type": "host",
            "type_id": 1,
            "children": [
                1,
                3,
                5,
                2,
                4,
                0
            ]
        },
        {
            "id": 0,
            "name": "osd.0",
            "type": "osd",
            "type_id": 0,
            "crush_weight": 0.002991,
            "depth": 2,
            "exists": 1,
            "status": "up",
            "reweight": 1.000000,
            "primary_affinity": 1.000000
        },
        {
            "id": 4,
            "name": "osd.4",
            "type": "osd",
            "type_id": 0,
            "crush_weight": 0.002991,
            "depth": 2,
            "exists": 1,
            "status": "down",
            "reweight": 1.000000,
            "primary_affinity": 1.000000
        },
        {
            "id": 2,
            "name": "osd.2",
            "type": "osd",
            "type_id": 0,
            "crush_weight": 0.002991,
            "depth": 2,
            "exists": 1,
            "status": "up",
            "reweight": 1.000000,
            "primary_affinity": 1.000000
        },
        {
            "id": 5,
            "name": "osd.5",
            "type": "osd",
            "type_id": 0,
            "crush_weight": 0.002991,
            "depth": 2,
            "exists": 1,
            "status": "up",
            "reweight": 1.000000,
            "primary_affinity": 1.000000
        },
        {
            "id": 3,
            "name": "osd.3",
            "type": "osd",
            "type_id": 0,
            "crush_weight": 0.002991,
            "depth": 2,
            "exists": 1,
            "status": "up",
            "reweight": 1.000000,
            "primary_affinity": 1.000000
        },
        {
            "id": 1,
            "name": "osd.1",
            "type": "osd",
            "type_id": 0,
            "crush_weight": 0.002991,
            "depth": 2,
            "exists": 1,
            "status": "up",
            "reweight": 1.000000,
            "primary_affinity": 1.000000
        }
    ],
    "stray": []
}
""".strip()

CEPH_REPORT = """
report 1188805303
{
    "cluster_fingerprint": "0d5f8f7a-0241-4a2e-8401-9dcf37a1039b",
    "version": "12.2.8-52.el7cp",
    "commit": "3af3ca15b68572a357593c261f95038d02f46201",
    "timestamp": "2019-06-05 23:33:08.514032",
    "tag": "",
    "health": {
        "checks": {
            "POOL_APP_NOT_ENABLED": {
                "severity": "HEALTH_WARN",
                "summary": {
                    "message": "application not enabled on 1 pool(s)"
                },
                "detail": [
                    {
                        "message": "application not enabled on pool 'pool-a'"
                    },
                    {
                        "message": "use 'ceph osd pool application enable <pool-name> <app-name>', where <app-name> is 'cephfs', 'rbd', 'rgw', or freeform for custom applications."
                    }
                ]
            },
            "MON_DOWN": {
                "severity": "HEALTH_WARN",
                "summary": {
                    "message": "1/3 mons down, quorum ceph2,ceph_1"
                },
                "detail": [
                    {
                        "message": "mon.ceph3 (rank 0) addr 10.72.37.76:6789/0 is down (out of quorum)"
                    }
                ]
            }
        },
        "status": "HEALTH_WARN",
        "summary": [
            {
                "severity": "HEALTH_WARN",
                "summary": "'ceph health' JSON format has changed in luminous. If you see this your monitoring system is scraping the wrong fields. Disable this with 'mon health preluminous compat warning = false'"
            }
        ],
        "overall_status": "HEALTH_WARN",
        "detail": [
            "'ceph health' JSON format has changed in luminous. If you see this your monitoring system is scraping the wrong fields. Disable this with 'mon health preluminous compat warning = false'"
        ]
    },
    "monmap_first_committed": 1,
    "monmap_last_committed": 1,
    "quorum": [
        1,
        2
    ],
    "osdmap_first_committed": 1,
    "osdmap_last_committed": 92
}
""".strip()

CEPH_REPORT_INVALID_JSON = """
report 1188805303
    "cluster_fingerprint": "0d5f8f7a-0241-4a2e-8401-9dcf37a1039b",
    "version": "12.2.8-52.el7cp",
    "commit": "3af3ca15b68572a357593c261f95038d02f46201",
    "timestamp": "2019-06-05 23:33:08.514032",
    "tag": "",
    "health": {
        "checks": {
            "POOL_APP_NOT_ENABLED": {
                "severity": "HEALTH_WARN",
                "summary": {
                    "message": "application not enabled on 1 pool(s)"
                },
                "detail": [
                    {
                        "message": "application not enabled on pool 'pool-a'"
                    },
                    {
                        "message": "use 'ceph osd pool application enable <pool-name> <app-name>', where <app-name> is 'cephfs', 'rbd', 'rgw', or freeform for custom applications."
                    }
                ]
            },
            "MON_DOWN": {
                "severity": "HEALTH_WARN",
                "summary": {
                    "message": "1/3 mons down, quorum ceph2,ceph_1"
                },
                "detail": [
                    {
                        "message": "mon.ceph3 (rank 0) addr 10.72.37.76:6789/0 is down (out of quorum)"
                    }
                ]
            }
        },
        "status": "HEALTH_WARN",
        "summary": [
            {
                "severity": "HEALTH_WARN",
                "summary": "'ceph health' JSON format has changed in luminous. If you see this your monitoring system is scraping the wrong fields. Disable this with 'mon health preluminous compat warning = false'"
            }
        ],
        "overall_status": "HEALTH_WARN",
        "detail": [
            "'ceph health' JSON format has changed in luminous. If you see this your monitoring system is scraping the wrong fields. Disable this with 'mon health preluminous compat warning = false'"
        ]
    },
    "monmap_first_committed": 1,
    "monmap_last_committed": 1,
    "quorum": [
        1,
        2
    ],
    "osdmap_first_committed": 1,
    "osdmap_last_committed": 92
}
""".strip()

CEPH_REPORT_INVALID = """
1188805303
{
    "cluster_fingerprint": "0d5f8f7a-0241-4a2e-8401-9dcf37a1039b",
    "version": "12.2.8-52.el7cp",
    "commit": "3af3ca15b68572a357593c261f95038d02f46201",
    "timestamp": "2019-06-05 23:33:08.514032",
    "tag": "",
    "health": {
        "checks": {
            "POOL_APP_NOT_ENABLED": {
                "severity": "HEALTH_WARN",
                "summary": {
                    "message": "application not enabled on 1 pool(s)"
                },
                "detail": [
                    {
                        "message": "application not enabled on pool 'pool-a'"
                    },
                    {
                        "message": "use 'ceph osd pool application enable <pool-name> <app-name>', where <app-name> is 'cephfs', 'rbd', 'rgw', or freeform for custom applications."
                    }
                ]
            },
            "MON_DOWN": {
                "severity": "HEALTH_WARN",
                "summary": {
                    "message": "1/3 mons down, quorum ceph2,ceph_1"
                },
                "detail": [
                    {
                        "message": "mon.ceph3 (rank 0) addr 10.72.37.76:6789/0 is down (out of quorum)"
                    }
                ]
            }
        },
        "status": "HEALTH_WARN",
        "summary": [
            {
                "severity": "HEALTH_WARN",
                "summary": "'ceph health' JSON format has changed in luminous. If you see this your monitoring system is scraping the wrong fields. Disable this with 'mon health preluminous compat warning = false'"
            }
        ],
        "overall_status": "HEALTH_WARN",
        "detail": [
            "'ceph health' JSON format has changed in luminous. If you see this your monitoring system is scraping the wrong fields. Disable this with 'mon health preluminous compat warning = false'"
        ]
    },
    "monmap_first_committed": 1,
    "monmap_last_committed": 1,
    "quorum": [
        1,
        2
    ],
    "osdmap_first_committed": 1,
    "osdmap_last_committed": 92
}
""".strip()


def test_ceph_doc_examples():
    env = {
        'ceph_osd_dump': CephOsdDump(context_wrap(CEPH_OSD_DUMP_INFO)),
        'ceph_osd_df': CephOsdDf(context_wrap(CEPH_OSD_DF_INFO)),
        'ceph_s': CephS(context_wrap(CEPH_S_INFO)),
        'ceph_df_detail': CephDfDetail(context_wrap(CEPH_DF_DETAIL_INFO)),
        'ceph_health_detail': CephHealthDetail(context_wrap(CEPH_HEALTH_DETAIL_INFO)),
        'ceph_osd_ec_profile_get': CephECProfileGet(context_wrap(CEPH_OSD_EC_PROFILE_GET)),
        'ceph_cfg_info': CephCfgInfo(context_wrap(CEPHINFO)),
        'ceph_osd_tree': CephOsdTree(context_wrap(CEPH_OSD_TREE)),
        'ceph_report_content': CephReport(context_wrap(CEPH_REPORT)),
    }
    failed, total = doctest.testmod(ceph_cmd_json_parsing, globs=env)
    assert failed == 0


class TestCephOsdDump():
    def test_ceph_osd_dump(self):
        result = CephOsdDump(context_wrap(CEPH_OSD_DUMP_INFO))

        assert result.data == {
            'pool_max': 12, 'max_osd': 8,
            'created': '2016-11-12 16:08:46.307206',
            'modified': '2017-03-07 08:55:53.301911',
            'epoch': 210, 'flags': u'sortbitwise',
            'cluster_snapshot': '',
            'fsid': '2734f9b5-2013-48c1-8e96-d31423444717',
            'pools': [
                {
                    'pool_name': 'rbd', 'flags_names': 'hashpspool',
                    'min_size': 2, 'object_hash': 2, 'flags': 1,
                    'pg_num': 256, 'crush_ruleset': 0, 'type': 1,
                    'pool': 0, 'size': 3
                }
            ]
        }
        assert result['pools'][0]['min_size'] == 2

    def test_ceph_osd_dump_empty(self):
        assert 'Empty output.' in skip_exception_check(CephOsdDump)


class TestCephOsdDf():
    def test_ceph_osd_df(self):
        result = CephOsdDf(context_wrap(CEPH_OSD_DF_INFO))

        assert result.data == {
            "nodes": [
                {
                    "id": 0,
                    "name": "osd.0",
                    "type": "osd",
                    "type_id": 0,
                    "crush_weight": 1.091095,
                    "depth": 2,
                    "reweight": 1.000000,
                    "kb": 1171539620,
                    "kb_used": 4048208,
                    "kb_avail": 1167491412,
                    "utilization": 0.345546,
                    "var": 1.189094,
                    "pgs": 945
                }
            ],
            "stray": [],
            "summary": {
                "total_kb": 8200777340,
                "total_kb_used": 23831128,
                "total_kb_avail": 8176946212,
                "average_utilization": 0.290596,
                "min_var": 0.803396,
                "max_var": 1.189094,
                "dev": 0.035843
            }
        }
        assert result['nodes'][0]['pgs'] == 945

    def test_ceph_os_df_empty(self):
        assert 'Empty output.' in skip_exception_check(CephOsdDf)


class TestCephS():
    def test_ceph_s(self):
        result = CephS(context_wrap(CEPH_S_INFO))

        assert result.data == {
            "health": {

            },
            "pgmap": {
                "pgs_by_state": [
                    {
                        "state_name": "active+clean",
                        "count": 1800
                    }
                ],
                "version": 314179,
                "num_pgs": 1800,
                "data_bytes": 7943926574,
                "bytes_used": 24405610496,
                "bytes_avail": 8373190385664,
                "bytes_total": 8397595996160
            },
            "fsmap": {
                "epoch": 1,
                "by_rank": []
            }
        }
        assert result['pgmap']['pgs_by_state'][0]['state_name'] == 'active+clean'

    def test_ceph_s_empty(self):
        assert 'Empty output.' in skip_exception_check(CephS)


class TestCephECProfileGet():
    def test_ceph_ec_profile_get(self):
        result = CephECProfileGet(context_wrap(CEPH_OSD_EC_PROFILE_GET))

        assert result.data == {
            "k": "2",
            "m": "1",
            "plugin": "jerasure",
            "technique": "reed_sol_van"
        }
        assert result['k'] == "2"
        assert result['m'] == "1"

    def test_ceph_ec_profile_get_empty(self):
        assert 'Empty output.' in skip_exception_check(CephECProfileGet)


class TestCephCfgInfo():
    def test_ceph_cfg_info(self):
        result = CephCfgInfo(context_wrap(CEPHINFO))

        assert result.data == {
            'ms_tcp_nodelay': 'true', 'ms_max_backoff': '15',
            'cluster': 'ceph', 'ms_dispatch_throttle_bytes': '104857600',
            'debug_none': '0/5', 'ms_crc_data': 'true', 'perf': 'true',
            'ms_tcp_prefetch_max_size': '4096', 'ms_die_on_bad_msg': 'false',
            'ms_bind_port_max': '7300', 'ms_bind_port_min': '6800',
            'ms_die_on_skipped_message': 'false', 'heartbeat_file': '',
            'heartbeat_interval': '5', 'heartbeat_inject_failure': '0',
            'ms_crc_header': 'true', 'max_open_files': '131072',
            'ms_die_on_old_message': 'false', 'name': 'osd.1',
            'ms_type': 'simple', 'ms_initial_backoff': '0.2',
            'ms_bind_retry_delay': '5', 'ms_bind_ipv6': 'false',
            'ms_die_on_unhandled_msg': 'false', 'ms_tcp_rcvbuf': '0',
            'ms_bind_retry_count': '3'
        }

        assert result.max_open_files == '131072'

    def test_ceph_cfg_info_empty(self):
        assert 'Empty output.' in skip_exception_check(CephCfgInfo)


class TestCephHealthDetail():
    def test_ceph_health_detail(self):
        result = CephHealthDetail(context_wrap(CEPH_HEALTH_DETAIL_INFO))

        assert result.data == {
            "health": {
            },
            "timechecks": {
                "epoch": 4,
                "round": 0,
                "round_status": "finished"
            },
            "summary": [],
            "overall_status": "HEALTH_OK",
            "detail": []
        }
        assert result['overall_status'] == 'HEALTH_OK'

    def test_ceph_health_detail_empty(self):
        assert 'Empty output.' in skip_exception_check(CephHealthDetail)


class TestCephDfDetail():
    def test_ceph_df_detail(self):
        result = CephDfDetail(context_wrap(CEPH_DF_DETAIL_INFO))

        assert result.data == {
            "stats": {
                "total_bytes": 17113243648,
                "total_used_bytes": 203120640,
                "total_avail_bytes": 16910123008,
                "total_objects": 0
            },
            "pools": [
                {
                    "name": "rbd",
                    "id": 0,
                    "stats": {
                        "kb_used": 0,
                        "bytes_used": 0,
                        "max_avail": 999252180,
                        "objects": 0,
                        "dirty": 0,
                        "rd": 0,
                        "rd_bytes": 0,
                        "wr": 0,
                        "wr_bytes": 0,
                        "raw_bytes_used": 0
                    }
                },
                {
                    "name": "ecpool",
                    "id": 2,
                    "stats": {
                        "kb_used": 0,
                        "bytes_used": 0,
                        "max_avail": 1998504360,
                        "objects": 0,
                        "dirty": 0,
                        "rd": 0,
                        "rd_bytes": 0,
                        "wr": 0,
                        "wr_bytes": 0,
                        "raw_bytes_used": 0
                    }
                }
            ]
        }
        assert result['stats']['total_avail_bytes'] == 16910123008

    def test_ceph_df_detail_empty(self):
        assert 'Empty output.' in skip_exception_check(CephDfDetail)


class TestCephOsdTree():
    def test_ceph_osd_tree(self):
        result = CephOsdTree(context_wrap(CEPH_OSD_TREE))
        assert result.data == {
            "nodes": [
                {
                    "id": -1,
                    "name": "default",
                    "type": "root",
                    "type_id": 10,
                    "children": [
                        -5,
                        -4,
                        -3,
                        -2
                    ]
                },
                {
                    "id": -2,
                    "name": "dhcp-192-56",
                    "type": "host",
                    "type_id": 1,
                    "children": []
                },
                {
                    "id": -3,
                    "name": "dhcp-192-104",
                    "type": "host",
                    "type_id": 1,
                    "children": []
                },
                {
                    "id": -4,
                    "name": "dhcp-192-67",
                    "type": "host",
                    "type_id": 1,
                    "children": []
                },
                {
                    "id": -5,
                    "name": "localhost",
                    "type": "host",
                    "type_id": 1,
                    "children": [
                        1,
                        3,
                        5,
                        2,
                        4,
                        0
                    ]
                },
                {
                    "id": 0,
                    "name": "osd.0",
                    "type": "osd",
                    "type_id": 0,
                    "crush_weight": 0.002991,
                    "depth": 2,
                    "exists": 1,
                    "status": "up",
                    "reweight": 1.000000,
                    "primary_affinity": 1.000000
                },
                {
                    "id": 4,
                    "name": "osd.4",
                    "type": "osd",
                    "type_id": 0,
                    "crush_weight": 0.002991,
                    "depth": 2,
                    "exists": 1,
                    "status": "down",
                    "reweight": 1.000000,
                    "primary_affinity": 1.000000
                },
                {
                    "id": 2,
                    "name": "osd.2",
                    "type": "osd",
                    "type_id": 0,
                    "crush_weight": 0.002991,
                    "depth": 2,
                    "exists": 1,
                    "status": "up",
                    "reweight": 1.000000,
                    "primary_affinity": 1.000000
                },
                {
                    "id": 5,
                    "name": "osd.5",
                    "type": "osd",
                    "type_id": 0,
                    "crush_weight": 0.002991,
                    "depth": 2,
                    "exists": 1,
                    "status": "up",
                    "reweight": 1.000000,
                    "primary_affinity": 1.000000
                },
                {
                    "id": 3,
                    "name": "osd.3",
                    "type": "osd",
                    "type_id": 0,
                    "crush_weight": 0.002991,
                    "depth": 2,
                    "exists": 1,
                    "status": "up",
                    "reweight": 1.000000,
                    "primary_affinity": 1.000000
                },
                {
                    "id": 1,
                    "name": "osd.1",
                    "type": "osd",
                    "type_id": 0,
                    "crush_weight": 0.002991,
                    "depth": 2,
                    "exists": 1,
                    "status": "up",
                    "reweight": 1.000000,
                    "primary_affinity": 1.000000
                }
            ],
            "stray": []
        }

        assert len(result['nodes'][0]['children']) == 4

    def test_ceph_osd_tree_empty(self):
        assert 'Empty output.' in skip_exception_check(CephOsdTree)


class TestCephReport():
    def test_ceph_report(self):
        result = CephReport(context_wrap(CEPH_REPORT))
        assert result['version'] == "12.2.8-52.el7cp"

    def test_invalid(self):
        with pytest.raises(ParseException) as e:
            CephReport(context_wrap(CEPH_REPORT_INVALID))
        assert "Invalid" in str(e)

    def test_invalid_json(self):
        with pytest.raises(ParseException) as e:
            CephReport(context_wrap(CEPH_REPORT_INVALID_JSON))
        assert "Could not parse json." in str(e)

    def test_ceph_report_empty(self):
        assert 'Empty output.' in skip_exception_check(CephReport)



from insights.tests import deep_compare


def test_deep_compare():
    """
    Tests that deep_compare() doesn't modify the objects it compares.
    """
    x_list = [8, 5, 3, 1, 2, 9, 5]
    x_list_backup = [8, 5, 3, 1, 2, 9, 5]
    b_tuple = (8, 5, 3, 1, 2, 9, 5)
    b_tuple_backup = (8, 5, 3, 1, 2, 9, 5)
    a = {"4d": 20, "5a": 10, "3b": 15, "7c": 5, "x": x_list,
         "b": b_tuple}
    b = {"4d": 20, "5a": 10, "3b": 15, "7c": 5, "x": [8, 5, 3, 1, 2, 9, 5],
         "b": (8, 5, 3, 1, 2, 9, 5)}
    c = {"4d": 20, "5a": 10, "3b": 15, "7c": 5, "x": [8, 5, 3, 1, 2, 9, 5],
         "b": (8, 5, 3, 1, 2, 9, 5)}
    d = {"4d": 20, "5a": 10, "3b": 15, "7c": 5, "x": [0, 8, 5, 3, 1, 2, 9, 5],
         "b": (8, 5, 3, 1, 2, 9, 5)}
    # sanity test before any manipulation
    # python can compare dicts natively to this extent
    assert c != d
    assert a == b
    assert a == c
    assert b == c
    assert x_list == x_list_backup
    assert b_tuple == b_tuple_backup
    noexception = False
    try:
        deep_compare(a, b)
        noexception = True
    except:
        pass
    # deep_compare test - the original objects should stay unchanged
    assert noexception
    assert a == b
    assert a == c
    assert b == c
    assert x_list == x_list_backup
    assert b_tuple == b_tuple_backup



"""
ReadLink parsers for Openshift certificate symbolic file links
==============================================================

This module contains the following parsers:

ReadLinkEKubeletClientCurrent - command ``/usr/bin/readlink -e /etc/origin/node/certificates/kubelet-client-current.pem``
-------------------------------------------------------------------------------------------------------------------------
ReadLinkEKubeletServerCurrent - command ``/usr/bin/readlink -e /etc/origin/node/certificates/kubelet-server-current.pem``
-------------------------------------------------------------------------------------------------------------------------
"""

from insights.specs import Specs
from insights.parsers import SkipException
from insights import parser, CommandParser


@parser(Specs.readlink_e_shift_cert_client)
class ReadLinkEKubeletClientCurrent(CommandParser):
    """
    Class for command: /usr/bin/readlink -e /etc/origin/node/certificates/kubelet-client-current.pem

    Sample content from command is::

        /etc/origin/node/certificates/kubelet-client-2019-10-18-23-17-35.pem

    Examples:
        >>> client.path
        '/etc/origin/node/certificates/kubelet-client-2019-10-18-23-17-35.pem'

    Raises:
        SkipException: When input content is empty
    """
    def parse_content(self, content):
        if content is None or len(content) == 0:
            raise SkipException("No Data from command: /usr/bin/readlink -e /etc/origin/node/certificates/kubelet-client-current.pem")

        self._path = content[-1]

    @property
    def path(self):
        """Returns real file path of /etc/origin/node/certificates/kubelet-client-current.pem"""
        return self._path


@parser(Specs.readlink_e_shift_cert_server)
class ReadLinkEKubeletServerCurrent(CommandParser):
    """
    Class for command: /usr/bin/readlink -e /etc/origin/node/certificates/kubelet-server-current.pem

    Sample content from command is::

        /etc/origin/node/certificates/kubelet-server-2018-10-18-23-29-14.pem

    Examples:
        >>> server.path
        '/etc/origin/node/certificates/kubelet-server-2018-10-18-23-29-14.pem'

    Raises:
        SkipException: When input content is empty
    """
    def parse_content(self, content):
        if content is None or len(content) == 0:
            raise SkipException("No Data from command: /usr/bin/readlink -e /etc/origin/node/certificates/kubelet-server-current.pem")

        self._path = content[-1]

    @property
    def path(self):
        """Returns real file path of /etc/origin/node/certificates/kubelet-server-current.pem"""
        return self._path



import json
from insights.parsr.examples.json_parser import (TRUE, FALSE, NULL, JsonArray,
        JsonObject, JsonValue)

DATA0 = """
{
    "name": "Adventure Lookup"
}
""".strip()


DATA1 = """
{
    "name": "Adventure Lookup",
    "icons": [
        {
            "src": "/android-chrome-192x192.png",
            "sizes": "192x192",
            "type": "image/png"
        },
        {
            "src": "/android-chrome-512x512.png",
            "sizes": "512x512",
            "type": "image/png"
        }
    ],
    "theme_color": "#ffffff",
    "background_color": "#ffffff",
    "display": "standalone"
}
""".strip()

DATA2 = """
{
  "meta": {
    "version": "0.6.0"
  },
  "GROUPS": {
    "c-development": {
      "grp_types": 0,
      "ui_name": "C Development Tools and Libraries",
      "name": "C Development Tools and Libraries",
      "full_list": [
        "valgrind",
        "automake",
        "indent",
        "autoconf",
        "ltrace",
        "bison",
        "ccache",
        "gdb",
        "strace",
        "elfutils",
        "byacc",
        "oprofile",
        "gcc-c++",
        "pkgconfig",
        "binutils",
        "gcc",
        "libtool",
        "cscope",
        "ctags",
        "flex",
        "glibc-devel",
        "make"
      ],
      "pkg_exclude": [],
      "pkg_types": 6
    }
  },
  "ENVIRONMENTS": {}
}""".strip()


def test_true():
    assert TRUE("true") is True


def test_false():
    assert FALSE("false") is False


def test_null():
    assert NULL("null") is None


def test_json_value_number():
    assert JsonValue("123") == 123


def test_json_value_string():
    assert JsonValue('"key"') == "key"


def test_json_empty_array():
    assert JsonArray("[]") == []


def test_json_single_element_array():
    assert JsonArray("['key']") == ["key"]


def test_json_multi_element_array():
    assert JsonArray("[1, 2, 3]") == [1, 2, 3]
    assert JsonArray("['key', -3.4, 'thing']") == ["key", -3.4, "thing"]


def test_json_nested_array():
    assert JsonArray("[1, [4, 5], 3]") == [1, [4, 5], 3]
    assert JsonArray("['key', [-3.4], 'thing']") == ["key", [-3.4], "thing"]


def test_json_empty_object():
    assert JsonObject("{}") == {}


def test_json_single_object():
    assert JsonObject('{"key": "value"}') == {"key": "value"}


def test_json_multi_object():
    expected = {"key1": "value1", "key2": 15}
    assert JsonObject('{"key1": "value1", "key2": 15}') == expected


def test_json_nested_object():
    text = '{ "key1": ["value1", "value2"], "key2": {"num": 15, "num2": 17 }}'
    expected = {"key1": ["value1", "value2"], "key2": {"num": 15, "num2": 17}}
    assert JsonObject(text) == expected


def test_data0():
    expected = json.loads(DATA0)
    assert JsonValue(DATA0) == expected


def test_data1():
    expected = json.loads(DATA1)
    assert JsonValue(DATA1) == expected


def test_data2():
    expected = json.loads(DATA2)
    assert JsonValue(DATA2) == expected



"""
LVDisplay - command ``/sbin/lvdisplay``
=======================================

The normal lvdisplay content looks like this::

  Adding lvsapp01ap01:0 as an user of lvsapp01ap01_mlog
  --- Volume group ---
  VG Name               vgp01app
  ...
  VG Size               399.98 GiB
  VG UUID               JVgCxE-UY84-C0Gk-8Cmn-UGXu-UHo0-9Qa4Re
  --- Logical volume ---
  global/lvdisplay_shows_full_device_path not found in config: defaulting to 0
  LV Path                /dev/vgp01app/lvsapp01ap01-old
  LV Name                lvsapp01ap01-old
  ...
  VG Name                vgp01app
  --- Logical volume ---
  global/lvdisplay_shows_full_device_path not found in config: defaulting to 0
  LV Path                /dev/vgp01app/lvsapp01ap02
  LV Name                lvsapp01ap02
  ...
  VG Name                vgp01app

The data is compiled into two keys in the ``data`` attribute:

* ``Logical volume``: a list of logical volume dictionaries.
* ``Volume group``: a list of volume group dictionaries.

The keys in each dictionary correspond to the headings found in the
output - for example, the keys in each ``Volume group`` list entry will
include ``VG Name``, ``VG Size``, etc.

In addition, the ``debug`` key in both the ``data`` attribute dictionary
and the ``Logical volume`` and ``Volume group`` dictionaries stores any
debug or warning messages found while parsing the output for that section.

Logical volumes are also available as a dictionary in the ``lvs`` property
and volume groups in the ``vgs`` property, both arranged by name.  Both
contain the same information as the associated list entry in the ``volumes``
dictionary.

Examples:

    >>> lvs = shared(LvDisplay)
    >>> 'volumes' in lvs # direct access via LegacyItemAccess
    True
    >>> 'debug' in lvs.data['volumes'] # access via data property
    True
    >>> for lv in lvs.data['volumes']['Logical volume']:
    ---     print lv['LV Name']
    ---
    lvsapp01ap01-old
    lvsapp01ap02
    >>> lvs.lvs['lvsapp01ap02']['VG Name'] # access to LVs by name
    'vgp01app'
    >>> lvs.vgs['vgp01app']['VG Size'] # access to VGs by name
    '399.98 GiB'

"""

from .. import parser, LegacyItemAccess, CommandParser
from collections import defaultdict
import re
from insights.specs import Specs


@parser(Specs.lvdisplay)
class LvDisplay(CommandParser, LegacyItemAccess):
    """
    Read the output of ``/sbin/lvdisplay``.

    Attributes:
        data(dict): The full data parsed from the output of lvdisplay.
        lvs(dict): A dictionary of logical volumes by name.
        vgs(dict): A dictionary of volume groups by name.
    """

    def parse_content(self, content):
        self.data = {}
        segment = []
        segment_type = ''
        self.data['debug'] = []
        self.data['volumes'] = defaultdict(list)
        for line in content:
            split_line = line.split()
            if segment and len(split_line) >= 2 and split_line[0] == '---' and split_line[-1] == '---':
                self.add_segment(segment_type, segment)
                segment_type = " ".join(split_line[1:len(split_line) - 1])
                segment = []
            else:
                segment.append(line)

        if segment:
            # last segment
            self.add_segment(segment_type, segment)

        self.data['volumes'] = dict(self.data['volumes'])

        # Add lvs and vgs properties with dicts from the data collected
        if 'Volume group' in self.data['volumes']:
            self.vgs = {}
            for vg in self.data['volumes']['Volume group']:
                self.vgs[vg['VG Name']] = vg
        if 'Logical volume' in self.data['volumes']:
            self.lvs = {}
            for lv in self.data['volumes']['Logical volume']:
                self.lvs[lv['LV Name']] = lv

    def add_segment(self, segment_type, segment):
        schema = ()
        for line in segment:
            if line.lstrip().startswith('VG Name'):
                indexes = [(m.start(), m.end()) for m in re.finditer(r"\ +", line)]
                schema = (indexes[0][1], indexes[2][1])
                break

        if not schema:
            self.data['debug'].extend(segment)
        else:
            self.data['volumes'][segment_type].append(self.parse_segment(segment, schema))

    def parse_segment(self, lines, schema):
        segment = {}
        debug = []
        for line in lines:
            name = line[schema[0]:schema[1]]
            if name.startswith(' '):
                debug.append(line[schema[0]:])
            else:
                value = line[schema[1]:].strip()
                name = name.strip()
                if not name and not value:
                    continue
                segment[name] = value

        segment['debug'] = debug
        return segment



"""
OpenShift Get commands
======================

``oc get`` command is from openshift used to list the usage info - e.g.
pods info, dc info, services info, etc.  This shared parser is used to parse
``oc get XX --all-namespaces -o yaml`` command information. Parameters
"--all-namespaces" means collecting information from all projects and "-o yaml"
means the output is in YAML format.

Parsers included in this module are:

OcGetBc - command ``oc get bc -o yaml --all-namespaces``
--------------------------------------------------------

OcGetBuild - command ``oc get build -o yaml --all-namespaces``
--------------------------------------------------------------

OcGetDc - command ``oc get dc -o yaml --all-namespaces``
--------------------------------------------------------

OcGetEgressNetworkPolicy - command ``oc get egressnetworkpolicy -o yaml --all-namespaces``
------------------------------------------------------------------------------------------

OcGetEndPoints - command ``oc get endpoints -o yaml --all-namespaces``
----------------------------------------------------------------------

OcGetEvent - command ``oc get event -o yaml --all-namespaces``
--------------------------------------------------------------

OcGetNode - command ``oc get nodes -o yaml``
--------------------------------------------

OcGetPod - command ``oc get pod -o yaml --all-namespaces``
----------------------------------------------------------

OcGetProject - command ``oc get project -o yaml --all-namespaces``
------------------------------------------------------------------

OcGetPv - command ``oc get pv -o yaml --all-namespaces``
--------------------------------------------------------

OcGetPvc - command ``oc get pvc -o yaml --all-namespaces``
----------------------------------------------------------

OcGetRc - command ``oc get rc -o yaml --all-namespaces``
--------------------------------------------------------

OcGetRole - command ``oc get role -o yaml --all-namespaces``
------------------------------------------------------------

OcGetRolebinding - command ``oc get rolebinding -o yaml --all-namespaces``
--------------------------------------------------------------------------

OcGetRoute - command ``oc get route -o yaml --all-namespaces``
--------------------------------------------------------------

OcGetService - command ``oc get service -o yaml --all-namespaces``
------------------------------------------------------------------

OcGetConfigmap - command ``oc get configmap -o yaml --all-namespaces``
----------------------------------------------------------------------

Examples:
    >>> type(setting_dic)
    <class 'insights.parsers.openshift_get.OcGetService'>
    >>> setting_dic.data['items'][0]['kind']
    'Service'
    >>> setting_dic.data['items'][0]['spec']['clusterIP']
    '172.30.0.1'
    >>> setting_dic.data['items'][0]['metadata']['name']
    'kubernetes'
    >>> setting_dic.data['items'][1]['metadata']['name']
    'router-1'
    >>> "zjj" in setting_dic.data['items'][1]['metadata']['namespace']
    True
"""

from .. import YAMLParser, parser, CommandParser
from insights.specs import Specs


def metadata_name_items(data):
    return dict((item['metadata']['name'], item) for item in data['items'])


@parser(Specs.oc_get_bc)
class OcGetBc(CommandParser, YAMLParser):
    """Class to parse ``oc get bc -o yaml --all-namespaces``"""

    @property
    def build_configs(self):
        """ dict: Returns a dictionary of openshift build configs information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_build)
class OcGetBuild(YAMLParser):
    """Class to parse ``oc get build -o yaml --all-namespaces``"""

    @property
    def started_builds(self):
        """ dict: Returns a dictionary of openshift started build information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_dc)
class OcGetDc(CommandParser, YAMLParser):
    """Class to parse ``oc get dc -o yaml --all-namespaces``"""

    @property
    def deployment_configs(self):
        """ dict: Returns a dictionary of openshift deploymentconfigs information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_egressnetworkpolicy)
class OcGetEgressNetworkPolicy(CommandParser, YAMLParser):
    """Class to parse ``oc get egressnetworkpolicy -o yaml --all-namespaces``"""

    @property
    def egress_network_policies(self):
        """ dict: Returns a dictionary of openshift egress network policy information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_endpoints)
class OcGetEndPoints(CommandParser, YAMLParser):
    """Class to parse ``oc get endpoints -o yaml --all-namespaces``"""

    @property
    def endpoints(self):
        """ dict: Returns a dictionary of openshift endpoints information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_event)
class OcGetEvent(CommandParser, YAMLParser):
    """Class to parse ``oc get event -o yaml --all-namespaces``"""

    @property
    def events(self):
        """ dict: Returns a dictionary of openshift events information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_node)
class OcGetNode(CommandParser, YAMLParser):
    """Class to parse ``oc get nodes -o yaml``"""

    @property
    def nodes(self):
        """ dict: Returns a dictionary of openshift nodes information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_pod)
class OcGetPod(CommandParser, YAMLParser):
    """Class to parse ``oc get pod -o yaml --all-namespaces``"""

    @property
    def pods(self):
        """ dict: Returns a dictionary of openshift pods information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_project)
class OcGetProject(CommandParser, YAMLParser):
    """Class to parse ``oc get project -o yaml --all-namespaces``"""

    @property
    def projects(self):
        """ dict: Returns a dictionary of openshift project information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_pv)
class OcGetPv(CommandParser, YAMLParser):
    """Class to parse ``oc get pv -o yaml --all-namespaces``"""

    @property
    def persistent_volumes(self):
        """ dict: Returns a dictionary of openshift persistent volume information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_pvc)
class OcGetPvc(CommandParser, YAMLParser):
    """Class to parse ``oc get pvc -o yaml --all-namespaces``"""

    @property
    def persistent_volume_claims(self):
        """ dict: Returns a dictionary of openshift persistent volume claim information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_rc)
class OcGetRc(CommandParser, YAMLParser):
    """Class to parse ``oc get rc -o yaml --all-namespaces``"""

    @property
    def replication_controllers(self):
        """ dict: Returns a dictionary of openshift replication controllers information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_role)
class OcGetRole(CommandParser, YAMLParser):
    """Class to parse ``oc get role -o yaml --all-namespaces``"""

    @property
    def roles(self):
        """ dict: Returns a dictionary of openshift role information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_rolebinding)
class OcGetRolebinding(CommandParser, YAMLParser):
    """Class to parse ``oc get rolebinding -o yaml --all-namespaces``"""

    @property
    def rolebindings(self):
        """ dict: Returns a dictionary of openshift rolebind information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_route)
class OcGetRoute(CommandParser, YAMLParser):
    """Class to parse ``oc get route -o yaml --all-namespaces``"""

    @property
    def routes(self):
        """ dict: Returns a dictionary of openshift route information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_service)
class OcGetService(CommandParser, YAMLParser):
    """Class to parse ``oc get service -o yaml --all-namespaces``"""

    @property
    def services(self):
        """ dict: Returns a dictionary of openshift services information."""
        return metadata_name_items(self.data)


@parser(Specs.oc_get_configmap)
class OcGetConfigmap(CommandParser, YAMLParser):
    """Class to parse ``oc get configmap -o yaml --all-namespaces``"""

    @property
    def configmaps(self):
        """ dict: Returns a dictionary of openshift configmaps information."""
        return metadata_name_items(self.data)



import doctest
from insights.parsers import avc_hash_stats
from insights.parsers.avc_hash_stats import AvcHashStats
from insights.tests import context_wrap

AVC_HASH_STATS = """
entries: 509
buckets used: 290/512
longest chain: 7
""".strip()


def test_avc_hash_stats():
    hash_stats = avc_hash_stats.AvcHashStats(context_wrap(AVC_HASH_STATS))
    assert hash_stats.entries == 509
    assert hash_stats.buckets == 512
    assert hash_stats.buckets_used == 290
    assert hash_stats.longest_chain == 7


def test_avc_hash_stats_doc_examples():
    env = {
        'avc_hash_stats': AvcHashStats(
            context_wrap(AVC_HASH_STATS)),
    }
    failed, total = doctest.testmod(avc_hash_stats, globs=env)
    assert failed == 0



"""
HttpdV - command ``httpd -V``
=============================

Module for parsing the output of command ``httpd -V``.   The bulk of the
content is split on the colon and keys are kept as is.  Lines beginning with
'-D' are kept in a dictionary keyed under 'Server compiled with'; each
compilation option is a key in this sub-dictionary.  The value of the
compilation options is the value after the equals sign, if one is present,
or the value in brackets after the compilation option, or 'True' if only the
compilation option is present.
"""

from .. import parser, LegacyItemAccess, CommandParser
from insights.specs import Specs
from insights.parsers import SkipException


@parser(Specs.httpd_V)
class HttpdV(LegacyItemAccess, CommandParser):
    """
    Class for parsing ``httpd -V`` command output.

    The data is kept in the ``data`` property and can be accessed through the
    object itself thanks to the ``LegacyItemAccess`` parser class.

    Typical output of command ``httpd -V`` looks like::

        Server version: Apache/2.2.6 (Red Hat Enterprise Linux)
        Server's Module Magic Number: 20120211:24
        Compiled using: APR 1.4.8, APR-UTIL 1.5.2
        Architecture:   64-bit
        Server MPM:     Prefork
        Server compiled with....
        -D APR_HAS_SENDFILE
        -D APR_HAVE_IPV6 (IPv4-mapped addresses enabled)
        -D AP_TYPES_CONFIG_FILE="conf/mime.types"
        -D SERVER_CONFIG_FILE="conf/httpd.conf"

    Examples:
        >>> type(hv)
        <class 'insights.parsers.httpd_V.HttpdV'>
        >>> hv.mpm
        'prefork'
        >>> hv["Server's Module Magic Number"]
        '20120211:24'
        >>> hv['Server compiled with']['APR_HAS_SENDFILE']
        True
        >>> hv['Server compiled with']['APR_HAVE_IPV6']
        'IPv4-mapped addresses enabled'
        >>> hv['Server compiled with']['SERVER_CONFIG_FILE']
        'conf/httpd.conf'

    Attributes:
        data (dict): The bulk of the content is split on the colon and keys are
            kept as is.  Lines beginning with '-D' are kept in a dictionary
            keyed under 'Server compiled with'; each compilation option is a key
            in this sub-dictionary.  The value of the compilation options is the
            value after the equals sign, if one is present, or the value in
            brackets after the compilation option, or 'True' if only the
            compilation option is present.

    Raises:
        SkipException: When input content is empty or there is no parsed data.
    """

    def parse_content(self, content):
        if not content:
            raise SkipException("Input content is empty.")

        self.data = {}
        compiled_with = {}
        for line in content:
            line = line.strip()
            if ': ' in line:
                key, value = [s.strip() for s in line.split(': ', 1)]
                self.data[key] = value.lower()
            elif line.startswith('-'):
                line = line[3:]  # cut off the '-D '
                if '=' in line:
                    key, value = line.split('=', 1)
                    compiled_with[key] = value.strip('"')
                else:
                    if ' ' in line:
                        key, value = line.split(' ', 1)
                        value = value.strip('"()')
                    else:
                        key, value = line, True
                    compiled_with[key] = value

        if self.data:
            if compiled_with:
                self.data['Server compiled with'] = compiled_with
        else:
            raise SkipException("Input content is not empty but there is no useful parsed data.")

    @property
    def httpd_command(self):
        """
        Return the full binary path of a running httpd or None when nothing
        is found. It's to identify which httpd binaries the instance run with.
        """
        return self.args

    @property
    def mpm(self):
        """
        str: The MPM mode of the running httpd. An Empty string when nothing
        is found.
        """
        return self.data.get('Server MPM', '')

    @property
    def version(self):
        """
        str: The version of the running httpd. An Empty string when nothing
        is found.
        """
        return self.get('Server version', '')
